//---------------------------------------------------------------------------
#include "CGShader.h"

//---------------------------------------------------------------------------

CGShader::CGShader(bool vertex_shader) : Shader(vertex_shader) {
	this->cgProfile = CG_PROFILE_UNKNOWN;
	//this->cgProfile = vertex_shader ? g3->getGraphicsEnvironment()->cgVertexProfile : g3->getGraphicsEnvironment()->cgPixelProfile;
	this->cgProgram = NULL;
}

CGShader::~CGShader() {
	if( this->cgProgram != NULL ) {
		cgDestroyProgram(cgProgram);
		this->cgProgram = NULL;
	}
}

void CGShader::setMatSpecular(const float specular[4]) const {
	/*if( cgparam_SpecMat != NULL ) {
		cgGLSetParameter4fv(cgparam_SpecMat, specular);
	}*/
	this->SetUniformParameter4fv("specular", specular);
}

void CGShader::setLightPos(float p1, float p2, float p3, float d1, float d2, float d3) const {
	/*if( cgparam_LightPos != NULL )
		cgGLSetParameter4f( cgparam_LightPos, p1, p2, p3, 0.0f );
	if( cgparam_LightVec != NULL )
		cgGLSetParameter4f( cgparam_LightVec, d1, d2, d3, 0.0f );*/
	this->SetUniformParameter4f("lt.LightPos", p1, p2, p3, 0.0f);
	this->SetUniformParameter4f("lightpos", p1, p2, p3, 0.0f);
	this->SetUniformParameter4f("lt.LightVec", d1, d2, d3, 0.0f);
}

void CGShader::setLightAmbient(float rf, float gf, float bf) const {
	/*if( cgparam_AmbLight != NULL )
		cgGLSetParameter4f( cgparam_AmbLight, rf, gf, bf, 1.0f );*/
	this->SetUniformParameter4f("lt.AmbLight", rf, gf, bf, 1.0f);
}

void CGShader::setLightDiffuse(float rf, float gf, float bf) const {
	/*if( cgparam_DiffLight != NULL )
		cgGLSetParameter4f( cgparam_DiffLight, rf, gf, bf, 1.0f );*/
	this->SetUniformParameter4f("lt.DiffLight", rf, gf, bf, 1.0f);
}

void CGShader::setLightSpecular(float rf, float gf, float bf) const {
	/*if( cgparam_SpecLight != NULL )
		cgGLSetParameter4f( cgparam_SpecLight, rf, gf, bf, 1.0f );*/
	this->SetUniformParameter4f("lt.SpecLight", rf, gf, bf, 1.0f);
}

void CGShader::setLightAtten(float a, float b, float c) const {
	/*if( cgparam_AmbLight != NULL )
		cgGLSetParameter4f( cgparam_AttenLight, a, b, c, 0.0f );*/
	this->SetUniformParameter3f("lt.AttenLight", a, b, c);
}

void CGShader::setBumpTexture(const Texture *texture) const {
	/*const GLTexture *gl_texture = static_cast<const GLTexture *>(texture);
	if( cgparam_BumpTextureID != 0 ) {
		cgGLSetParameter1f( cgparam_BumpTextureID, gl_texture != NULL ? (float)gl_texture->getTextureID() : 0.0f );
		//cgSetParameter1i( cgparam_BumpTextureID, texture != NULL ? 1 : 0 );
	}*/
	this->SetUniformParameter1f("bump_textureID", texture != NULL ? 1.0f : 0.0f);
}

void CGShader::setHardwareAnimationAlpha(float alpha) const {
	/*if( cgparam_frame_alpha != 0 ) {
		cgGLSetParameter1f( cgparam_frame_alpha, alpha );
	}*/
	SetUniformParameter1f("frame_alpha", alpha);
}
