//---------------------------------------------------------------------------

#include "D3D9GraphicsEnvironment.h"
#include "D3D9Renderer.h"
#include "SoftwareRenderer.h"
#include "../GraphicsEngine.h"
#include "../Texture.h"
#include "../Win32/Win32InputHandler.h"

#include <dxerr.h>
#include <Cg/CgD3D9.h>

#pragma comment(lib, "dxerr.lib")

//---------------------------------------------------------------------------

//D3D9GraphicsEnvironment::D3D9GraphicsEnvironment(int width,int height,int multisample,bool fullscreen,bool hdr) : GraphicsEnvironment(width, height, multisample, fullscreen) {
//D3D9GraphicsEnvironment::D3D9GraphicsEnvironment(VisionPrefs *visionPrefs) : GraphicsEnvironment(visionPrefs->resolution_width, visionPrefs->resolution_height, visionPrefs->multisample, visionPrefs->fullscreen) {
D3D9GraphicsEnvironment::D3D9GraphicsEnvironment(VisionPrefs visionPrefs) : GraphicsEnvironment(visionPrefs), hWnd(NULL) {
	//const bool software = false;
	//const bool software = true;
	bool software = visionPrefs.api == VISIONAPI_SOFTWARE;
	//const bool want_shaders = false;
	//const bool want_shaders = true;
	bool want_shaders = visionPrefs.shaders;
	LOG("startup Direct3D 9 GraphicsEnvironment\n");
	RendererInfo::init();

	this->active = false; // needs to be initialised before we start the D3D device!
	ZeroMemory( &d3dpp, sizeof(d3dpp) );
	this->m_bDeviceLost = false;
	//this->m_bDeviceLost = true;

	HINSTANCE hInstance = GetModuleHandle(NULL);

	this->cgContext = NULL;
	this->cgVertexProfile = CG_PROFILE_UNKNOWN;
	this->cgPixelProfile = CG_PROFILE_UNKNOWN;
	this->m_pD3D = NULL;
	this->m_pDevice = NULL;

	// Register the window class.
	/*WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, D3D9GraphicsEnvironment::WndProc, 0L, 0L, 
		GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
		TEXT("Vision3_main"), NULL };*/
	/*
	// original
	WNDCLASS wc;
	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style =  CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = ( WNDPROC )D3D9GraphicsEnvironment::WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = TEXT("Vision3_main");

	//if( !RegisterClassEx( &wc ) ) {
	if( !RegisterClass( &wc ) ) {
		LOG("Failed to register class\n");
		return;
	}*/

	WNDCLASSEX wc;
	memset(&wc, 0, sizeof(wc));
	wc.cbSize = sizeof(wc);
	wc.style = CS_HREDRAW | CS_VREDRAW;
	//wc.lpfnWndProc = StaticWndProc;
	wc.lpfnWndProc = D3D9GraphicsEnvironment::WndProc;
	wc.hInstance = hInstance;
	//wc.hIcon = LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
	//wc.hCursor = LoadCursor(NULL, MAKEINTRESOURCE(IDC_ARROW));
	wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	//wc.lpszClassName = s_szWindowClassname;
	wc.lpszClassName = TEXT("Vision3_main");
	wc.hIconSm = wc.hIcon;
	if( !RegisterClassEx(&wc) ) {
		Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_OPEN_WINDOW,"Failed to register class"));
		return;
	}

	/*const int bits = 32;
	if( visionPrefs.fullscreen ) {
		DEVMODE dmScreenSettings;
		memset(&dmScreenSettings,0,sizeof(dmScreenSettings));
		dmScreenSettings.dmSize=sizeof(dmScreenSettings);
		dmScreenSettings.dmPelsWidth  = visionPrefs.resolution_width;
		dmScreenSettings.dmPelsHeight = visionPrefs.resolution_height;
		dmScreenSettings.dmBitsPerPel = bits;
		dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;
		// Try To Set Selected Mode And Get Results.  NOTE: CDS_FULLSCREEN Gets Rid Of Start Bar.
		if (ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL)
		{
			//Vision::setError(new VisionException(this,VisionException::V_DISPLAYMODE_NOT_AVAILABLE,"The requested displaymode was not available"));
			LOG("The requested displaymode was not available\n");
			closeScreen();
			return;
		}
	}*/

	DWORD dwExStyle = 0;
	DWORD dwStyle = 0;
	RECT WindowRect;

	WindowRect.left=(long)0;
	WindowRect.right=(long)visionPrefs.resolution_width;
	WindowRect.top=(long)0;
	WindowRect.bottom=(long)visionPrefs.resolution_height;

	if ( visionPrefs.fullscreen ) {
		dwExStyle=WS_EX_APPWINDOW;
		//dwStyle = NULL;
		dwStyle=WS_POPUP;
		//ShowCursor(FALSE);
	}
	else {
		dwExStyle=WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		//dwStyle=WS_OVERLAPPEDWINDOW;
		//dwStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;
		dwStyle = WS_OVERLAPPEDWINDOW & ~WS_THICKFRAME & ~WS_MAXIMIZEBOX;
		/*if( software )
			dwStyle = dwStyle | WS_THICKFRAME | WS_MAXIMIZEBOX;*/
	}
	AdjustWindowRectEx(&WindowRect, dwStyle, FALSE, dwExStyle); // Adjust Window To True Requested Size

	// Create the application's window.
	LOG("about to create window\n");
	/*hWnd = CreateWindow( TEXT("Vision3_main"), TEXT("Vision3 (c) 2000-2009 Mark Harman"),
		WS_OVERLAPPEDWINDOW, 0, 0, width, height,
		GetDesktopWindow(), NULL, wc.hInstance, NULL );
	if( hWnd == NULL ) {*/
	if( !( hWnd = CreateWindowEx( dwExStyle,		// Extended Style For The Window
		TEXT("Vision3_main"),						// Class Name
		TEXT("Vision 3 window"),	// Window Title
		//WS_CLIPSIBLINGS |							// Required Window Style
		//WS_CLIPCHILDREN |							// Required Window Style
		dwStyle,									// Selected Window Style
		0, 0,										// Window Position
		WindowRect.right-WindowRect.left,			// Calculate Adjusted Window Width
		WindowRect.bottom-WindowRect.top,			// Calculate Adjusted Window Height
		NULL,										// No Parent Window
		NULL,										// No Menu
		hInstance,									// Instance
		NULL ) ) )					 				// Don't Pass Anything To WM_CREATE
	{
		Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_OPEN_WINDOW,"Could not create window"));
		return;
	}
	LOG("created window OK\n");

	m_pD3D = Direct3DCreate9( D3D_SDK_VERSION );
	LOG("D3D %x\n", m_pD3D);
	if( m_pD3D == NULL ) {
		Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_INIT_3D,"Could not create Direct3D 9 object"));
		return;
	}
	LOG("initialised Direct3D 9 OK\n");
	D3DADAPTER_IDENTIFIER9 adapterIdentifier;
	m_pD3D->GetAdapterIdentifier(D3DADAPTER_DEFAULT, 0, &adapterIdentifier);
	LOG("D3D Driver: %s\n", adapterIdentifier.Driver);
	LOG("D3D Description: %s\n", adapterIdentifier.Description);
	LOG("D3D Device Name: %s\n", adapterIdentifier.DeviceName);
	LOG("D3D Vendor ID: %d\n", adapterIdentifier.VendorId);
	LOG("D3D Device ID: %d\n", adapterIdentifier.DeviceId);
	LOG("D3D SubSys ID: %d\n", adapterIdentifier.SubSysId);
	LOG("D3D Revision: %d\n", adapterIdentifier.Revision);
	LOG("D3D WHQLLevel: %d\n", adapterIdentifier.WHQLLevel);
	LOG("D3D Product: %d\n", HIWORD(adapterIdentifier.DriverVersion.HighPart));
	LOG("D3D Version: %d\n", LOWORD(adapterIdentifier.DriverVersion.HighPart));
	LOG("D3D SubVersion: %d\n", HIWORD(adapterIdentifier.DriverVersion.LowPart));
	LOG("D3D Build: %d\n", LOWORD(adapterIdentifier.DriverVersion.LowPart));

	//D3DFORMAT backBufferFormat = visionPrefs.fullscreen ? D3DFMT_A8R8G8B8 : D3DFMT_UNKNOWN;
	D3DFORMAT backBufferFormat = D3DFMT_A8R8G8B8;
	D3DFORMAT depthstencilFormat = D3DFMT_D24S8;
	bool multisample_available = false;
	D3DMULTISAMPLE_TYPE multisample_type = D3DMULTISAMPLE_NONE;
	if( visionPrefs.multisample == 0 ) {
		// no multisampling
	}
	else if( visionPrefs.multisample == 2 )
		multisample_type = D3DMULTISAMPLE_2_SAMPLES;
	else if( visionPrefs.multisample == 4 )
		multisample_type = D3DMULTISAMPLE_4_SAMPLES;
	else if( visionPrefs.multisample == 8 )
		multisample_type = D3DMULTISAMPLE_8_SAMPLES;
	else {
		Vision::setError(new VisionException(this, VisionException::V_PREFS_ERROR, "prefs: unrecognised multisample value requested"));
	}
	DWORD multisample_levels = 0;

	if( visionPrefs.multisample == 0 ) {
		LOG("multisampling not requested\n");
	}
	else if( SUCCEEDED( m_pD3D->CheckDeviceMultiSampleType( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, backBufferFormat, !visionPrefs.fullscreen, multisample_type, &multisample_levels ) ) ) {
		if( SUCCEEDED( m_pD3D->CheckDeviceMultiSampleType( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, depthstencilFormat, !visionPrefs.fullscreen, multisample_type, &multisample_levels ) ) ) {
			multisample_available = true;
			LOG("multisampling available: %d levels\n", multisample_levels);
		}
		else {
			LOG("no multisampling - incompatible depthstencil buffer:(\n");
		}
	}
	else {
		LOG("no multisampling - incompatible back buffer:(\n");
	}

	D3DCAPS9 caps;
	m_pD3D->GetDeviceCaps( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps );
	LOG("obtained device caps OK\n");
	if( caps.MaxActiveLights <= 0 ) {
		// this apparently implies no limit (due to being done in software)
		// set an arbitrary limit that's good enough
		RendererInfo::maxLights = 8;
		LOG("No limit on max Lights (software mode) - set max equal to %d\n", RendererInfo::maxLights);
	}
	else {
		RendererInfo::maxLights = caps.MaxActiveLights;
		LOG("Max Lights? %d\n", RendererInfo::maxLights);
	}

	if( caps.StencilCaps & D3DSTENCILCAPS_TWOSIDED ) {
		D3D9RendererInfo::twoSidedStencil = true;
	}
	LOG("Two sided stencil? %s\n", D3D9RendererInfo::twoSidedStencil?"T":"F");
	if( D3D9RendererInfo::twoSidedStencil && !visionPrefs.twosidedstencil ) {
		D3D9RendererInfo::twoSidedStencil = false;
		LOG(">>> Disabling two sided stencil\n");
	}

	if( caps.Caps2 & D3DCAPS2_CANAUTOGENMIPMAP ) {
		LOG("can autogenerate mipmaps\n");
	}
	else {
		LOG("can't autogenerate mipmaps\n");
	}

	RendererInfo::framebufferObject = true; // TODO: assume true for now?

	d3dpp.Windowed = !visionPrefs.fullscreen;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = backBufferFormat;
	d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE; // no vsync
	d3dpp.EnableAutoDepthStencil = TRUE;
	//d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	d3dpp.AutoDepthStencilFormat = depthstencilFormat;
	d3dpp.hDeviceWindow = hWnd;
	d3dpp.BackBufferCount = 1;
	d3dpp.BackBufferWidth = visionPrefs.resolution_width;
	d3dpp.BackBufferHeight = visionPrefs.resolution_height;
	if( software ) {
		d3dpp.Flags = D3DPRESENTFLAG_LOCKABLE_BACKBUFFER; // needed for saveScreen() too? but can't use with multisampling!
	}
	if( multisample_available ) {
		d3dpp.MultiSampleType = multisample_type;
	}
	/*d3dpp.BackBufferWidth = WindowRect.right-WindowRect.left;
	d3dpp.BackBufferHeight = WindowRect.bottom-WindowRect.top;*/
	/*if( fullscreen ) {
		d3dpp.BackBufferWidth = WindowRect.right-WindowRect.left;
		d3dpp.BackBufferHeight = WindowRect.bottom-WindowRect.top;
	}
	else {
		// wrong!
		d3dpp.BackBufferWidth = GetSystemMetrics(SM_CXSCREEN);
		d3dpp.BackBufferHeight = GetSystemMetrics(SM_CYSCREEN);
	}*/

	DWORD behaviour;
	if ( caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT ) {
		LOG("hardware transform and lighting available\n");
		behaviour = D3DCREATE_HARDWARE_VERTEXPROCESSING;
		if( !visionPrefs.hardwaretl ) {
			behaviour = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
			LOG(">>> Disabling hardware transform and lighting\n");
		}
	}
	else {
		LOG("hardware transform and lighting not available :( using software mode\n");
		behaviour = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
	}

	this->m_pDevice = NULL;
	if( FAILED( m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
		behaviour,
		&d3dpp, &m_pDevice ) ) ) {
			Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_INIT_3D,"Could not create Direct3D 9 Device"));
			return;
	}
	LOG("D3D Device %x\n", m_pDevice);
	LOG("created device OK\n");

	this->g3 = new Graphics3D(this, visionPrefs.hdr);
	if( software ) {
		this->renderer = new SoftwareRenderer(this, m_pD3D, m_pDevice);
		this->g = new SoftwareGraphics2D(this, static_cast<SoftwareRenderer *>(renderer));
	}
	else {
		this->renderer = new D3D9Renderer(this, m_pD3D, m_pDevice);
		this->g = new D3D9Graphics2D(this, static_cast<D3D9Renderer *>(renderer));
	}
	this->g->owned = true; // owned by this GraphicsEnvironment class
	this->inputSystem = new InputSystem(new Win32MouseInputHandler(), new KeyboardInputHandler());
	LOG("created subsystems okay OK\n");

	if( want_shaders && !software ) {
		this->initCG();
	}
	g3->init(); // also inits shaders

	// we do this, as the code expects there to be a projection matrix (for the 3D viewing mode) on the stack
	this->renderer->switchToProjection();
	this->renderer->pushMatrix();
	this->renderer->switchToModelview();

	g3->ReSizeGLScene(renderer->getZNear(), renderer->getZFar(), visionPrefs.resolution_width, visionPrefs.resolution_height, renderer->getPersp(), fovy); // need to reset viewport/perspective info
	this->created = true;

	/*D3DMATERIAL9 mtrl;
	ZeroMemory( &mtrl, sizeof(mtrl) );
	mtrl.Diffuse.r = mtrl.Ambient.r = 1.0f;
	mtrl.Diffuse.g = mtrl.Ambient.g = 1.0f;
	mtrl.Diffuse.b = mtrl.Ambient.b = 1.0f;
	mtrl.Diffuse.a = mtrl.Ambient.a = 1.0f;
	m_pDevice->SetMaterial( &mtrl );

	m_pDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );
	m_pDevice->SetRenderState( D3DRS_LIGHTING, FALSE );

	// we want equivalent of glEnable(GL_COLOR_MATERIAL); glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	m_pDevice->SetRenderState( D3DRS_COLORVERTEX, TRUE ); // should be default
	m_pDevice->SetRenderState( D3DRS_DIFFUSEMATERIALSOURCE, D3DMCS_COLOR1 ); // should be default
	m_pDevice->SetRenderState( D3DRS_AMBIENTMATERIALSOURCE, D3DMCS_COLOR1 ); // *not* the default - by default is taken from material
	//m_pDevice->SetRenderState( D3DRS_DIFFUSEMATERIALSOURCE, D3DMCS_MATERIAL );
	//m_pDevice->SetRenderState( D3DRS_AMBIENTMATERIALSOURCE, D3DMCS_MATERIAL );*/

	this->multisample_ok = visionPrefs.multisample > 0; // if we're here and it was requested, it must have been successful
	/*if( !multisample_ok && visionPrefs.multisample > 0 ) {
		LOG("failed to create screen with multisampling\n");
		visionPrefs.multisample = 0; // reset multisample to 0 if not available
	}*/

	//this->m_bDeviceLost = false;
	LOG("all done!\n");
}

D3D9GraphicsEnvironment::~D3D9GraphicsEnvironment() {
	closeScreen();

	delete g;
	delete g3;
	delete renderer;
	delete inputSystem;

	if( cgContext != NULL ) {
		cgD3D9SetDevice(NULL); // needed to avoid hang on exit, on Intel GMA 950
		cgDestroyContext(cgContext);
		cgContext = NULL;
	}
}

void D3D9GraphicsEnvironment::closeScreen() {
	this->sound = NULL; // clear reference to block calls to it, as it may have been deleted by this stage

	if( m_pDevice != NULL ) {
		m_pDevice->Release();
		m_pDevice = NULL;
	}

	if( m_pD3D != NULL ) {
		m_pD3D->Release();
		m_pD3D = NULL;
	}

	/*if(visionPrefs.fullscreen) {
		ChangeDisplaySettings(NULL,0);
		//ShowCursor(TRUE);
	}*/
	if( hWnd != NULL ) {
		//ReleaseDC( hWnd, hDC );
		// destroy the window explicitly
		DestroyWindow( hWnd );
		hWnd = NULL;
	}

	if( !UnregisterClassA("Vision3_main", GetModuleHandle(NULL)) ) {
		LOG("could not unregister class!\n");
	}
}

void D3D9GraphicsEnvironment::cgErrorCallback(void) {
	CGerror error = cgGetError();
	LOG("CG ERROR %d : %s\n", error, cgGetErrorString(error));
	const char *lastError = cgGetLastErrorString(NULL);
	if( lastError != NULL ) {
		LOG("Last Error String: %s\n", lastError);
	}
	D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(GraphicsEnvironment::getSingleton());
	const char *lastListing = cgGetLastListing(genv->cgContext);
	if( lastListing != NULL ) {
		LOG("Last Listing String: %s\n", lastListing);
	}
	Vision::setError(new VisionException(NULL,VisionException::V_CG_ERROR,"CG reported an error"));
}

void D3D9GraphicsEnvironment::initCG() {
	// setup CG
	checkError("GraphicsEnvironment::initCG enter");

	cgSetErrorCallback(cgErrorCallback);

	cgContext = cgCreateContext();
	if( cgContext != NULL ) {
		LOG("Successfully initialised CG Context\n");

		cgD3D9SetDevice(m_pDevice);
		cgVertexProfile = cgD3D9GetLatestVertexProfile();
		if( cgVertexProfile == CG_PROFILE_UNKNOWN ) {
			LOG("### CG ERROR: Invalid vertex profile type\n");
			cgDestroyContext(cgContext);
			cgContext = NULL;
		}
		else {
			LOG("Successfully obtained vertex profile\n");
			//cgGLSetOptimalOptions(cgVertexProfile);

			cgPixelProfile = cgD3D9GetLatestPixelProfile();
			if( cgPixelProfile == CG_PROFILE_UNKNOWN ) {
				LOG("### CG ERROR: Invalid pixel profile type\n");
				cgDestroyContext(cgContext);
				cgContext = NULL;
			}
			else {
				LOG("Successfully obtained pixel profile\n");
				//cgGLSetOptimalOptions(cgPixelProfile);
			}
		}
	}
	else {
		LOG("### Failed to initialise CG Context\n");
	}

	//cgGLSetDebugMode(CG_FALSE);
	checkError("GraphicsEnvironment::initCG exit");
}

Shader *D3D9GraphicsEnvironment::createShader(const char *file, const char *func, bool vertex_shader) {
	if( !this->hasShaders() )
		return NULL;
	CGD3D9Shader *shader = new CGD3D9Shader(static_cast<D3D9Renderer *>(g3->getRenderer()), vertex_shader);
	LOG("Loading %s shader %d : %s / %s\n", vertex_shader?"vertex":"pixel", shader, file, func);
	shader->cgProfile = vertex_shader ? this->cgVertexProfile : this->cgPixelProfile;

	const char **profileOpts = cgD3D9GetOptimalOptions(shader->cgProfile);

	shader->cgProgram = cgCreateProgramFromFile(this->cgContext, CG_SOURCE, file, shader->cgProfile, func, profileOpts);
	if( shader->cgProgram == NULL ) {
		LOG("### Failed to load %s shader:\n", vertex_shader?"vertex":"pixel");
		CGerror Error = cgGetError();
		LOG(cgGetErrorString(Error));
		LOG("\n");
		delete shader;
		//this->cgContext = NULL;
		return NULL;
	}
	cgD3D9LoadProgram(shader->cgProgram, FALSE, 0);
	shader->init();
	g3->addShader(shader);

	return shader;
}

void D3D9GraphicsEnvironment::setVisible(bool visible) {
	if( visible ) {
		ShowWindow(hWnd,SW_SHOW);
		SetForegroundWindow(hWnd);
		SetFocus(hWnd);
	}
	else
		ShowWindow(hWnd,SW_HIDE);
}

void D3D9GraphicsEnvironment::setWindowTitle(const char *title) {
	SetWindowTextA(hWnd, title);
}

void D3D9GraphicsEnvironment::swapBuffers() {
	this->renderer->endScene();
	HRESULT hResult = this->m_pDevice->Present(NULL, NULL, NULL, NULL);
	if( FAILED( hResult ) ) {
		LOG("swapBuffers lost device\n");
		if( !handlePresentRetVal(hResult) ) {
			LOG("failed to recover!\n");
			Vision::setError(new VisionException(this, VisionException::V_DIRECT3D_LOSTDEVICE, "swapBuffers lost device"));
		}
		else {
			LOG("...recovered okay\n");
		}
	}
}

void D3D9GraphicsEnvironment::processEvents() {
	// most events handled by WndProc callback
	MSG msg;
	if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {
		// handle or dispatch messages
		if ( msg.message == WM_QUIT ) {
			request_quit = true;
		} else {
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
	}
}

void D3D9GraphicsEnvironment::waitForEvents() {
	// call Win32 API
	WaitMessage();
}

void D3D9GraphicsEnvironment::sleep(int time) {
	// call Win32 API
	Sleep(time);
}

bool D3D9GraphicsEnvironment::saveScreen(const char *filename,const char *type) {
	// we can't lock a surface unless we set the flag D3DPRESENTFLAG_LOCKABLE_BACKBUFFER - but that can't be used with multisampling
	// so we use D3DXSaveTextureToFile instead
	LPDIRECT3DSURFACE9 sourceSurface = NULL;
	if( FAILED( this->m_pDevice->GetBackBuffer(0, 0, D3DBACKBUFFER_TYPE_MONO, &sourceSurface) ) ) {
		LOG("failed to get back buffer\n");
		return false;
	}

	bool ok = false;
	D3DXIMAGE_FILEFORMAT format = D3DXIFF_PNG;
	if( stricmp(type, "bmp") )
		format = D3DXIFF_BMP;
	else if( stricmp(type, "jpg") )
		format = D3DXIFF_JPG;
	else if( stricmp(type, "tga") )
		format = D3DXIFF_TGA;
	else if( stricmp(type, "png") )
		format = D3DXIFF_PNG;
	else if( stricmp(type, "dds") )
		format = D3DXIFF_DDS;
	else if( stricmp(type, "ppm") )
		format = D3DXIFF_PPM;
	else if( stricmp(type, "dib") )
		format = D3DXIFF_DIB;
	else if( stricmp(type, "hdr") )
		format = D3DXIFF_HDR;
	else if( stricmp(type, "pfm") )
		format = D3DXIFF_PFM;
	else {
		Vision::setError(new VisionException(this, VisionException::V_IMAGE_UNSUPPORTED_FILEFORMAT, "saveScreen: unrecognised or unsupported file format requested\n"));
	}
	if( FAILED( D3DXSaveSurfaceToFileA(filename, format, sourceSurface, NULL, NULL) ) ) {
		LOG("failed to save surface\n");
	}
	else {
		ok = true;
	}
	sourceSurface->Release();
	return ok;

	/*bool ok = false;
	LPDIRECT3DSURFACE9 sourceSurface = NULL;
	this->m_pDevice->GetBackBuffer(0, 0, D3DBACKBUFFER_TYPE_MONO, &sourceSurface);

	D3DLOCKED_RECT d3drect;
	if( !FAILED( sourceSurface->LockRect(&d3drect, NULL, D3DLOCK_READONLY) ) ) {
		Image2D *image = new Image2D(this->getWidth(), this->getHeight());
		unsigned char *src = (unsigned char *)d3drect.pBits;
		unsigned char *dest = image->getData();
		for(int y=0;y<this->getHeight();y++) {
			unsigned char *line = &dest[y*d3drect.Pitch];
			for(int x=0;x<this->getWidth();x++) {
				unsigned char b = *src++;
				unsigned char g = *src++;
				unsigned char r = *src++;
				src++;
				*dest++ = r;
				*dest++ = g;
				*dest++ = b;
			}
		}
		sourceSurface->UnlockRect();
		ok = image->save(filename, type);
		delete image;
	}
	else {
		LOG("failed to lock screen\n");
	}
	sourceSurface->Release();
	return ok;*/
}

LRESULT CALLBACK D3D9GraphicsEnvironment::WndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam ) {
	switch ( message ) {
		case WM_ACTIVATE:
			{
				//D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getGenv(hWnd));
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
				if(genv != NULL) {
					genv->active = ( LOWORD( wParam ) != WA_INACTIVE );
					//LOG("    %d\n", genv->active);
				}
			}
			break;
		/*case WM_ACTIVATEAPP:
			{
				LOG("active:\n");
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getGenv(hWnd));
				if(genv!=NULL) {
					genv->active = ( LOWORD( wParam ) != WA_INACTIVE );
					LOG("    %d\n", genv->active);
				}
				return 0;
			}
		case WM_CREATE:
			return 0;*/
		case WM_CLOSE:
			{
				PostQuitMessage( 0 );
				return 0;
			}
		case WM_DESTROY:
			//return 0;
			break;
		case WM_KEYDOWN:
			{
				//LOG("KEYDOWN: %d\n", wParam);
				//InputHandler::keys[wParam] = true;
				//GraphicsEnvironment *genv = getGenv(hWnd);
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->keys[wParam] = true;
				genv->getInputSystem()->getKeyboard()->setKey(wParam, true);
				/*if(wParam == VK_ESCAPE)
				PostQuitMessage( 0 );*/
				return 0;
			}
		case WM_KEYUP:
			{
				//LOG("KEYUP: %d\n", wParam);
				//InputHandler::keys[wParam] = false;
				//GraphicsEnvironment *genv = getGenv(hWnd);
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->keys[wParam] = false;
				genv->getInputSystem()->getKeyboard()->setKey(wParam, false);
				return 0;
			}
		case WM_LBUTTONDOWN:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
				//LOG(">>>leftmousedown\n");
				//genv->getInputHandler()->mouseLChange(true);
				genv->getInputSystem()->getMouse()->mouseLChange(true);
				break;
			}
		case WM_MBUTTONDOWN:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->mouseMChange(true);
				genv->getInputSystem()->getMouse()->mouseMChange(true);
				break;
			}
		case WM_RBUTTONDOWN:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->mouseRChange(true);
				genv->getInputSystem()->getMouse()->mouseRChange(true);
				break;
			}
		case WM_LBUTTONUP:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
				//LOG(">>>leftmouseup\n");
				//genv->getInputHandler()->mouseLChange(false);
				genv->getInputSystem()->getMouse()->mouseLChange(false);
				break;
			}
		case WM_MBUTTONUP:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->mouseMChange(false);
				genv->getInputSystem()->getMouse()->mouseMChange(false);
				break;
			}
		case WM_RBUTTONUP:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->mouseRChange(false);
				genv->getInputSystem()->getMouse()->mouseRChange(false);
				break;
			}
		case WM_SIZE:
			{
				/*D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getGenv(hWnd));
				//GraphicsEnvironment::ReSizeGLScene(LOWORD(lParam),HIWORD(lParam));  // LoWord=Width, HiWord=Height
				if(genv!=NULL)
					genv->resize(LOWORD(lParam),HIWORD(lParam));  // LoWord=Width, HiWord=Height*/
				//return 0;
				break;
			}
			/*default:
			return DefWindowProc( hWnd, message, wParam, lParam );*/
	}

	if( message != NULL ) {
		//GraphicsEnvironment *genv = getGenv(hWnd);
		D3D9GraphicsEnvironment *genv = static_cast<D3D9GraphicsEnvironment *>(getSingleton());
		if( genv->getSound() != NULL ) {
			genv->getSound()->processEvents(message, wParam, lParam);
		}
	}

	return DefWindowProc( hWnd, message, wParam, lParam );
}

bool D3D9GraphicsEnvironment::drawFrameStart() {
	HRESULT hResult = m_pDevice->TestCooperativeLevel();
	if( FAILED(hResult) ) {
		/*if(!HandlePresentRetVal(hResult))
			m_bShouldExit = true;
		return;*/
		LOG("drawFrameStart lost device\n");
		if( !handlePresentRetVal(hResult) ) {
			LOG("failed to recover!\n");
			Vision::setError(new VisionException(this, VisionException::V_DIRECT3D_LOSTDEVICE, "drawFrameStart lost device"));
		}
		else {
			LOG("...recovered okay\n");
		}
		return false;
	}
	return true;
}

bool D3D9GraphicsEnvironment::handlePresentRetVal(HRESULT hResult) {
	if( hResult == D3DERR_DEVICELOST ) {
		LOG("D3DERR_DEVICELOST\n");
		// we now leave calls to onDeviceLost() to resetDevice() - otherwise we have the problem where the application may have created D3D9 objects after this point, but before we receive D3DERR_DEVICENOTRESET!
		/*if( !m_bDeviceLost ) {
			onDeviceLost();
			m_bDeviceLost = true;
		}*/
	}
	else if( hResult == D3DERR_DEVICENOTRESET ) {
		LOG("D3DERR_DEVICENOTRESET\n");
		if( !resetDevice() )
			return false;
		m_bDeviceLost = false;
	}
	else if( hResult == D3DERR_DRIVERINTERNALERROR ) {
		LOG("D3DERR_DRIVERINTERNALERROR\n");
		// DX Docs state that we should try resetting
		/*if( !resetDevice() )
			return false;*/
		// TODO:
		return false;
	}
	else if( FAILED(hResult) ) {
		LOG("Failed: %s\n", DXGetErrorString(hResult));
		return false;
	}

	return true;
}

bool D3D9GraphicsEnvironment::resetDevice() {
	LOG("D3D9GraphicsEnvironment::resetDevice()\n");
	// Make sure we call OnDeviceLost to free anything before device is reset
	if( !m_bDeviceLost ) {
		onDeviceLost();
	}

	// Reset device
	LOG("about to reset device...\n");
	HRESULT hResult = m_pDevice->Reset(&d3dpp);
	if( FAILED( hResult ) )	{
		LOG("Failed to reset! Error: %s\n", DXGetErrorString(hResult));
		return false;
	}
	LOG("done\n");

	// TODO:
	// Update window style in case we changed between windowed and fullscreen
	/*if(m_thePresentParams.Windowed)
	{
		SetWindowLong(m_hWnd, GWL_STYLE, WS_OVERLAPPEDWINDOW);
		SetWindowPos(m_hWnd, HWND_TOP, 0, 0, 0, 0,
			SWP_FRAMECHANGED | SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW);
	}
	else
	{
		SetWindowLong(m_hWnd, GWL_STYLE, WS_POPUPWINDOW);
		SetWindowPos(m_hWnd, HWND_TOP, 0, 0, 0, 0,
			SWP_FRAMECHANGED | SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW);
	}*/

	// Device is no longer lost
	onDeviceReset();
	// TODO:
	//SetupState();
	/*hResult = m_pDevice->TestCooperativeLevel();
	if( FAILED(hResult) ) {
		LOG("TestCooperativeLevel failed after reset!: %s\n", DXGetErrorString(hResult));
	}*/
	m_bDeviceLost = false;
	return true;
}

void D3D9GraphicsEnvironment::onDeviceLost() {
	LOG("D3D9GraphicsEnvironment::onDeviceLost()\n");
	// The device is lost, or is about to be reset. This is the time to Release() any resources
	// in the default pool, and call OnLostDevice() for any D3DX objects that need it.
	if( visionPrefs.api == VISIONAPI_SOFTWARE ) {
		return;
	}
	D3D9Renderer *d_renderer = static_cast<D3D9Renderer *>(renderer);
	d_renderer->onDeviceLost();
	LOG("D3D9GraphicsEnvironment::onDeviceLost() done\n");
}

void D3D9GraphicsEnvironment::onDeviceReset() {
	LOG("D3D9GraphicsEnvironment::onDeviceReset()\n");
	// The device has just been reset. This is the time to recreate() any resources in the default
	// pool, and call OnResetDevice() for any D3DX objects that need it.
	if( visionPrefs.api == VISIONAPI_SOFTWARE ) {
		return;
	}
	D3D9Renderer *d_renderer = static_cast<D3D9Renderer *>(renderer);
	//Vision::setError(new VisionException(this, VisionException::V_DIRECT3D_ERROR, "about to reset!"));
	d_renderer->onDeviceReset();
	LOG("D3D9GraphicsEnvironment::onDeviceReset() done\n");
}
