#pragma once

#include <vector>
using std::vector;
#include <string>
using std::string;

class FontBuffer;
class InputSystem;

#include "Resource.h"
#include "VisionIface.h"
#include "Misc.h"

//typedef void (PaintPanel) (GraphicsEnvironment *genv,void *data);

class KeyRepeatHandler {
	int time_repeat;
	int lasttime_input;
public:
	KeyRepeatHandler() : time_repeat(-1), lasttime_input(-1) {
	}

	bool isNewRepeat(bool keydown, int thistime);
};

class Panel : public VisionObject, public virtual VI_Panel {
protected:
	Color bgcolor;
	Color fgcolor;
	VI_Texture *texture;
	//bool opaque;
	Panel *parent;
	vector<Panel *> *children;
	VI_Panel_Paint *paintPanel;
	void *paintData;
	VI_Panel_Action *action;
	VI_Panel_Action *mouse_enter_action;
	VI_Panel_Action *mouse_exit_action;
	void *userData;
	bool has_shortcut;
	int shortcut;
	string popup_text;
	const FontBuffer *popup_font;
	bool show_popup;
	int popup_x, popup_y;
	int popup_width, popup_height;

	bool mouse_over;
	int last_event_time;
	bool enabled;
	bool visible;
	bool has_input_focus;
	int fixed_x,fixed_y,width,height;

	virtual void inputChildren(InputSystem *iS);
	virtual void paintChildren(Graphics2D *g);
public:
	V_Action *action_tag;
	V_Action *mouse_enter_action_tag;
	V_Action *mouse_exit_action_tag;

	int x,y;
	bool from_left;
	bool from_top;

	Panel();
	virtual ~Panel();
	virtual V_CLASS_t getClass() const { return V_CLASS_PANEL; };
	virtual bool isSubClass(V_CLASS_t v_class) {
		return ( v_class == V_CLASS_PANEL );
	}
	virtual int memUsage();

	virtual void input(InputSystem *iS);
	virtual void paint(Graphics2D *g);
	virtual void validate();
	virtual void setBackground(Color bg);
	virtual void setForeground(Color fg);
	//virtual void setOpaque(bool opaque);
	void drawPopup(Graphics2D *g);

	// interface
	virtual V_TAG_t getTag() const {
		return this->tag;
	}
	virtual const char *getName() const {
		return VisionObject::getName();
	}
	virtual void setName(const char *name) {
		VisionObject::setName(name);
	}

	virtual void setEnabled(bool enabled,bool do_children);
	virtual bool isEnabled() const {
		return this->enabled;
	}
	virtual void setVisible(bool visible) {
		this->visible = visible;
	}
	virtual bool isVisible() const {
		return this->visible;
	}
	virtual void setHasInputFocus(bool has_input_focus) {
		this->has_input_focus = has_input_focus;
	}

	virtual void setPopupText(const char *popup_text, const VI_Font *font);
	virtual void add(VI_Panel *panel,int x,int y);
	virtual void remove(VI_Panel *panel);
	virtual void removeAll();
	virtual void detach();
	virtual void setMeasureFrom(bool from_left, bool from_top) {
		this->from_left = from_left;
		this->from_top = from_top;
		this->validate();
	}
	virtual void setAction(VI_Panel_Action *action);
	virtual void setMouseEnterAction(VI_Panel_Action *mouse_enter_action) {
		this->mouse_enter_action = mouse_enter_action;
	}
	virtual void setMouseExitAction(VI_Panel_Action *mouse_exit_action) {
		this->mouse_exit_action = mouse_exit_action;
	}
	virtual void setPaintFunc(VI_Panel_Paint *paintPanel,void *paintData);
	virtual void setKeyShortcut(int shortcut) {
		this->has_shortcut = true;
		this->shortcut = shortcut;
	}
	virtual void disableKeyShortcut() {
		this->has_shortcut = false;
		this->shortcut = -1;
	}
	virtual void setForeground(float r,float g,float b);
	virtual void setForeground(float r,float g,float b,float a);
	virtual void setBackground(float r,float g,float b);
	virtual void setBackground(float r,float g,float b,float a);
	virtual void setBackgroundAlphai(int alpha);
	virtual void setTexture(VI_Texture *texture) {
		this->texture = texture;
	}
	virtual void setSize(int width, int height) {
		this->width = width;
		this->height = height;
		this->validate();
	}
	virtual void getSize(int *width, int *height) const {
		*width = this->width;
		*height = this->height;
	}
	virtual int getWidth() const {
		return width;
	}
	virtual int getHeight() const {
		return height;
	}
	virtual void setUserData(void *userData) {
		this->userData = userData;
	}
	virtual void *getUserData() const {
		return userData;
	}
	virtual VI_Panel *find(int fx, int fy);
};

class FlowPanel : public Panel {
	bool across;
	int size;
	int marginX,marginY;
public:
	FlowPanel(bool across,int size);
	virtual V_CLASS_t getClass() const { return V_CLASS_FLOWPANEL; };
	virtual bool isSubClass(V_CLASS_t v_class) {
		return ( v_class == V_CLASS_FLOWPANEL || Panel::isSubClass(v_class) );
	}

	virtual void add(VI_Panel *panel);
	virtual void validate();
};

class Button : public Panel, public virtual VI_Button {
	string text;
	const FontBuffer *font;
	bool border;

	virtual void init(const char *text,const FontBuffer *font);
public:
	//Button();
	Button(const char *text,const FontBuffer *font);
	virtual ~Button();
	virtual V_CLASS_t getClass() const { return V_CLASS_BUTTON; };
	virtual bool isSubClass(V_CLASS_t v_class) {
		return ( v_class == V_CLASS_BUTTON || Panel::isSubClass(v_class) );
	}
	virtual int memUsage() {
		int size = Panel::memUsage();
		size += text.capacity();
		return size;
	}

	virtual void input(InputSystem *iS);
	virtual void paint(Graphics2D *g);

	// interface
	virtual void setBorder(bool border) {
		this->border = border;
	}
};

class ImageButton : public Panel, public virtual VI_ImageButton {
	VI_Texture *texture;
public:
	ImageButton(VI_Texture *texture);
	ImageButton(VI_Texture *texture, int width, int height);
	virtual ~ImageButton();
	virtual V_CLASS_t getClass() const { return V_CLASS_IMAGEBUTTON; };
	virtual bool isSubClass(V_CLASS_t v_class) {
		return ( v_class == V_CLASS_IMAGEBUTTON || Panel::isSubClass(v_class) );
	}
	virtual int memUsage() {
		int size = Panel::memUsage();
		// todo:
		return size;
	}

	virtual void input(InputSystem *iS);
	virtual void paint(Graphics2D *g);
};

class Cycle : public Panel, public virtual VI_Cycle {
	/*char **entries;
	int n_entries;*/
	vector<string> entries;

	const FontBuffer *font;

	virtual void init(char **entries,const FontBuffer *font);
	void updateWidth(const char *entry);
public:
	int active;

	Cycle();
	Cycle(char **entries,const FontBuffer *font);
	virtual ~Cycle();
	virtual V_CLASS_t getClass() const { return V_CLASS_CYCLE; };
	virtual bool isSubClass(V_CLASS_t v_class) {
		return ( v_class == V_CLASS_CYCLE || Panel::isSubClass(v_class) );
	}
	virtual int memUsage() {
		int size = Panel::memUsage();
		//size += n_entries * sizeof(char *);
		// TODO: entries
		return size;
	}

	virtual void input(InputSystem *iS);
	virtual void paint(Graphics2D *g);

	// interface
	virtual int getActive() const {
		return this->active;
	}
	virtual int getNEntries() const {
		return this->entries.size();
	}
	virtual void addEntry(const char *entry);

	virtual void setActive(int active) {
		this->active = active;
	}
};

/*class ListBoxEntry {
public:
char *text;
ListBoxEntry(char *text);
virtual ~ListBoxEntry();
};*/

class Listbox : public Panel, public virtual VI_Listbox {
	/*char **entries;
	int n_entries;*/
	vector<string> entries;
	int position;
	//Vector *entries;

	KeyRepeatHandler keyRepeatHandler;
	const FontBuffer *font;
	int active;
public:

	Listbox(int width,int height,char **entries,const FontBuffer *font);
	Listbox(int width,int height,const string *entries,int n_entries,const FontBuffer *font);
	virtual ~Listbox() {
	}
	virtual V_CLASS_t getClass() const { return V_CLASS_LISTBOX; };
	virtual bool isSubClass(V_CLASS_t v_class) {
		return ( v_class == V_CLASS_LISTBOX || Panel::isSubClass(v_class) );
	}
	virtual int memUsage() {
		int size = Panel::memUsage();
		//size += n_entries * sizeof(char *);
		// TODO: entries
		return size;
	}

	virtual void input(InputSystem *iS);
	virtual void paint(Graphics2D *g);

	// interface
	virtual int getNEntries() const {
		return this->entries.size();
	}
	virtual int getActive() const {
		return this->active;
	}
	virtual void setActive(int active) {
		this->active = active;
	}
	virtual void reset(const string *entries, int n_entries);
	virtual void reset(char **entries);
};

class StringGadget : public Panel, public virtual VI_Stringgadget {
protected:
	//StringBuffer *text;
	string text;        
	const FontBuffer *font;
	bool read_only;
	/*int time_repeat;
	int lasttime_input;*/
	KeyRepeatHandler keyRepeatHandler;
	int max_len;
	VI_Panel_Action *enter_action;
	int cursor_blink_time_start;

	virtual void init(const FontBuffer *font,int width);
	static bool acceptKey(int key);
public:
	V_Action *enter_action_tag;

	StringGadget(const FontBuffer *font,int width);
	StringGadget();
	virtual ~StringGadget();
	virtual V_CLASS_t getClass() const { return V_CLASS_STRINGGADGET; };
	virtual bool isSubClass(V_CLASS_t v_class) {
		return ( v_class == V_CLASS_STRINGGADGET || Panel::isSubClass(v_class) );
	}
	virtual int memUsage() {
		int size = Panel::memUsage();
		// TODO: text
		return size;
	}

	virtual void input(InputSystem *iS);
	virtual void paint(Graphics2D *g);

	// interface
	virtual void setReadOnly(bool read_only) {
		this->read_only = read_only;
	}
	virtual void setMaxLen(int max_len) {
		this->max_len = max_len;
	}
	virtual void addText(const char *text);
	virtual void clear();
	virtual void getText(char **text,int *length) const;
	virtual const char *getText() const {
		return this->text.c_str();
	}
	virtual void setEnterAction(VI_Panel_Action *action);
};

class TextField : public StringGadget, public virtual VI_Textfield {
	//int cline_off;
	int cpos;
	//Vector *line_indxs;
	vector<int> *line_indxs;
	bool at_end;
public:
	TextField(const FontBuffer *font,int width,int height);
	TextField();
	virtual ~TextField();
	virtual V_CLASS_t getClass() const { return V_CLASS_TEXTFIELD; };
	virtual bool isSubClass(V_CLASS_t v_class) {
		return ( v_class == V_CLASS_TEXTFIELD || StringGadget::isSubClass(v_class) );
	}

	virtual void addText(const char *text);
	virtual void paint(Graphics2D *g);
	virtual void input(InputSystem *iS);

	// interface
	virtual void setCPos(int cpos) {
		this->cpos = cpos;
	}
};
