#pragma once

#include "VisionUtils.h"

class Color;
class Renderer;
class GraphicsEnvironment;
class Texture;
class VertexArray;

const int MAX_SPLATS = 4;

class RenderData {
	//friend unsigned __stdcall castShadow_thread(void *ptr);
	friend void castShadow_thread(void *ptr);

	//void renderArrays(GLenum mode,unsigned int vbo_this_indices, unsigned short *va_this_indices, int n_this_renderlength, int n_this_vertices);
	void renderCore(Renderer *renderer, GraphicsEnvironment *genv, int frame, int frame2, float frac, bool forced_color);
	//void doArrays(Graphics3D *g3);

	// Help data
	float *faces_planes; // size 4 * n_frames * n_polys
	int *faces_adj_faces; // size n_renderlength
	bool *faces_boundary; // size n_renderlength
	bool *edges_skip; // size 3 * n_polys; can we skip this edge, for the shadow extrusions?
	bool *visible_polys; // size n_polys; work data for is a poly visible
	unsigned short *shadow_edges; // temp buffer for which edges to extrude (NEW algorithm only)
	bool *shadow_vertices_used; // temp buffer for which vertices to extrude (NEW algorithm only)
	bool shadow_closed;
	void clearHelpData();

	// shadow volume vertex arrays
	bool need_shadows;
	//unsigned int vbo_shadow_vertices;
	VertexArray *vao_shadow_vertices;
	float *va_shadow_vertices;
	float *va_shadow_normals;
	//unsigned int vbo_shadow_indices;
	VertexArray *vao_shadow_indices;
	unsigned short *va_shadow_indices;
	int n_shadow_renderlength;
	void freeShadowData();
	void castShadowSetVisible(int frame, const Vector3D *lightpos, bool infinite, bool shader);
	int findShadowEdges();
	static void shadowExtrudeVertices_thread(void *ptr);
	void castShadowConstructVertices(int frame, int frame2, float frac, const Vector3D *lightpos, bool infinite);
	void castShadowConstructIndices(int n_shadow_edges);
	int castShadowConstruct(int frame, int frame2, float frac, const Vector3D *lightpos, bool infinite, bool shader);
	void renderShadow(Renderer *renderer, GraphicsEnvironment *genv, int n_shadow_vertices) const;

public:
	//bool using_vbos;

	/*enum Type {
	RENDER_MANUAL = 0,
	RENDER_DISPLAYLISTS,
	RENDER_VERTEXBUFFERS
	};
	//GLenum mode;
	Type type;*/
	int n_frames;
	/*GLuint *displayLists;
	int displayList_n_polys;
	int displayList_n_verts;*/

	Texture *texture;
	Texture *secondary_texture;
	int n_splats; // nb - if using splats, we can't have secondary or bump textures
	Texture *splat_textures[MAX_SPLATS]; // 0th splat is the alphamap
	float splat_detail_scale;
	Texture *bump_texture;

	unsigned int n_vertices;

	// vertex arrays
	//unsigned int vbo_vertices;
	VertexArray *vao_vertices;
	float *va_vertices;

	//unsigned int vbo_vertices_stream;
	VertexArray *vao_vertices_stream;
	float *va_vertices_stream;

	//unsigned int vbo_normals;
	VertexArray *vao_normals;
	float *va_normals;

	VertexArray *vao_tangents;
	float *va_tangents;

	//unsigned int vbo_texcoords;
	VertexArray *vao_texcoords;
	float *va_texcoords;

	//unsigned int vbo_texcoords_secondary;
	//float *va_texcoords_secondary;

	//unsigned int vbo_colors;
	VertexArray *vao_colors;
	unsigned char *va_colors;

	//unsigned int vbo_indices;
	VertexArray *vao_indices;
	unsigned short *va_indices;

	//unsigned short *vx_index;
	//int n_indices;
	int n_renderlength;
	int last_frame;
	float last_frac;

	bool solid;
	bool blending;
	bool blend_glow;
	bool alpha_channel;
	//bool splatting; // splatting means drawing on top of a base mesh with alpha blending
	// kind of like the blending flag, but it's draw in the opaque pipeline!
	unsigned char cols[4]; // specifies RGBA colour for vertex buffers if colour array not set
	float specular[4];

	void free();

	RenderData();
	~RenderData() {
		free();
	}
	virtual int memUsage();
	/*void initDisplayLists(int n_frames) {
	free();
	this->type = RENDER_DISPLAYLISTS;
	this->n_frames = n_frames;
	this->displayLists = new GLuint[n_frames];
	for(int i=0;i<n_frames;i++)
	this->displayLists[i] = 0;
	}*/
	void initVertexArrays(int n_frames,int n_vertices,bool colors,bool alpha_channel,bool normals,bool texcoords/*,bool texcoords_secondary = false*/) {
		//initDisplayLists(n_frames);
		free();
		//this->type = RENDER_VERTEXBUFFERS;
		this->n_frames = n_frames;
		this->n_vertices = n_vertices;
		int n_elements = n_frames * n_vertices;
		if( normals ) {
			this->va_normals = new float[3 * n_elements];
		}
		this->va_vertices = new float[3 * n_elements];
		this->va_vertices_stream = new float[3 * n_vertices];
		this->alpha_channel = alpha_channel;
		if( colors ) {
			this->va_colors = new unsigned char[(alpha_channel ? 4 : 3) * n_vertices];
		}
		if( texcoords ) {
			this->va_texcoords = new float[2 * n_vertices];
		}
		/*if( texcoords_secondary ) {
		this->va_texcoords_secondary = new float[2 * n_vertices];
		}*/
	}
	bool compatible(const RenderData *that) const;
	void freeVBOs();
	void simplify();
	void shareVertices();
	void compact();
	void combine(const RenderData *that);
	//static RenderData *createShadowMesh(const RenderData *rd, int n_rd);
	RenderData *copy(bool need_shadows,bool colors,bool normals,bool texcoords) const;
	void setBumpTexture(Texture *bump_texture);
	void prepare(bool need_shadows);
	void render(Renderer *renderer, GraphicsEnvironment *genv, int frame, int frame2, float frac, bool force_color, const Color *forced_color, unsigned char forced_alpha, bool lighting_pass);
	void castShadow(Renderer *renderer, GraphicsEnvironment *genv, int frame, int frame2, float frac, const Vector3D *lightpos, bool infinite, bool shader);
	int getNVertices() const {
		return n_vertices;
	}
	Vector3D getVertex(int frame,int indx) const {
		Vector3D vec;
		vec.x = this->va_vertices[3*n_vertices*frame + 3*indx + 0];
		vec.y = this->va_vertices[3*n_vertices*frame + 3*indx + 1];
		vec.z = this->va_vertices[3*n_vertices*frame + 3*indx + 2];
		return vec;
	};
	Vector3D getNormal(int frame, int indx) const {
		Vector3D vec;
		vec.x = this->va_normals[3*n_vertices*frame + 3*indx + 0];
		vec.y = this->va_normals[3*n_vertices*frame + 3*indx + 1];
		vec.z = this->va_normals[3*n_vertices*frame + 3*indx + 2];
		return vec;
	}
	Vector3D getTexCoord(int indx) const {
		Vector3D vec;
		vec.x = this->va_texcoords[2*indx + 0];
		vec.y = this->va_texcoords[2*indx + 1];
		vec.z = 0.0;
		return vec;
	};
	unsigned int getVaIndex(int index) const {
		if( this->va_indices == NULL )
			return index;
		return this->va_indices[index];
	}

	void setTangent(int frame, int indx, Vector3D *vec) {
		this->va_tangents[3*n_vertices*frame + 3*indx + 0] = vec->x;
		this->va_tangents[3*n_vertices*frame + 3*indx + 1] = vec->y;
		this->va_tangents[3*n_vertices*frame + 3*indx + 2] = vec->z;
	}
};
