//---------------------------------------------------------------------------
#include <cassert>
#include <cstdarg>

#include "Renderer.h"
#include "GraphicsEnvironment.h"

//---------------------------------------------------------------------------

bool RendererInfo::multiSample = false;
int RendererInfo::maxLights = 0;
//int RendererInfo::maxTextureUnits = 0;
bool RendererInfo::hardwareOcclusion = false;
bool RendererInfo::framebufferObject = false;

#if 0
void RendererInfo::initGLExtensions(GraphicsEnvironment *genv) {
	maxLights = 0;
	maxTextureUnits = 0;
	hardwareOcclusion = false;
	compiledVAs = false;
	vertexBufferObjects = false;
	drawRangeElements = false;

	// first we need to create a dummy window, before we can use glGetString
	/*WNDCLASS wc;
	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = CS_OWNDC;
	wc.lpfnWndProc = ( WNDPROC )WndProc_dummy;
	wc.hInstance = hInstance;
	//wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
	//wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	//wc.hbrBackground = (HBRUSH)GetStockObject( BLACK_BRUSH );
	wc.lpszClassName = "Vision3_dummy";
	RegisterClass( &wc );

	HWND hwnd = CreateWindow("Vision3_dummy", "Vision3_dummy",
	WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
	0, 0, 32, 32, 0, 0, hInstance, 0);
	if( hwnd == NULL ) {
	Vision::setError(new VisionException(NULL,VisionException::V_FAILED_TO_OPEN_WINDOW,"Could not create dummy window"));
	return;
	}
	HDC hDC = GetDC(hwnd);

	PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.cColorBits = 16;
	pfd.cDepthBits = 15;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;

	int iFormat = ChoosePixelFormat( hDC, &pfd );
	SetPixelFormat( hDC, iFormat, &pfd );
	HGLRC hRC = wglCreateContext( hDC );
	if(hRC == NULL) {
	DestroyWindow(hwnd);
	UnregisterClass("Vision3_dummy", hInstance);
	Vision::setError(new VisionException(NULL,VisionException::V_FAILED_TO_INIT_OPENGL,"Could not create dummy OpenGL context"));
	return;
	}
	if(!wglMakeCurrent( hDC, hRC )) {
	DestroyWindow(hwnd);
	UnregisterClass("Vision3_dummy", hInstance);
	Vision::setError(new VisionException(NULL,VisionException::V_FAILED_TO_INIT_OPENGL,"Could not make dummy OpenGL context current"));
	return;
	}*/

	// now we can grab the extension info
	PFNWGLGETEXTENSIONSSTRINGARBPROC glGetExtensionsStringARB =
		(PFNWGLGETEXTENSIONSSTRINGARBPROC)wglGetProcAddress("wglGetExtensionsStringARB");
	const char *wgl_extension = NULL;
	if( glGetExtensionsStringARB ) {
		wgl_extension = glGetExtensionsStringARB(genv->getHDC());
	}

	char *extension = (char *)glGetString(GL_EXTENSIONS);
	if(extension == NULL) {
		/*DestroyWindow(hwnd);
		UnregisterClass("Vision3_dummy", hInstance);*/
		return;
	}
	LOG("Extensions: %s\n", extension);
	if( wgl_extension != NULL ) {
		LOG("WGL Extensions: %s\n", wgl_extension);
	}

	if(wgl_extension != NULL && strstr(wgl_extension, "WGL_ARB_pixel_format")) {
		pixelFormat = true;
		wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC)wglGetProcAddress("wglChoosePixelFormatARB");
		//__wglewChoosePixelFormatARB(hdc, iattr, 0, 256, formats, &count);
	}
	LOG("WGL Pixel Format? %s\n", pixelFormat?"T":"F");

	if(wgl_extension != NULL && strstr(wgl_extension, "WGL_ARB_multisample")) {
		multiSample = true;
	}
	LOG("Multisample? %s\n", multiSample?"T":"F");

	/*if(wgl_extension != NULL && strstr(wgl_extension, "WGL_ARB_pbuffer") && strstr(wgl_extension, "WGL_ARB_render_texture")) {
	pbuffer = true;
	wglCreatePbufferARB    = (PFNWGLCREATEPBUFFERARBPROC)wglGetProcAddress("wglCreatePbufferARB");
	wglGetPbufferDCARB     = (PFNWGLGETPBUFFERDCARBPROC)wglGetProcAddress("wglGetPbufferDCARB");
	wglReleasePbufferDCARB = (PFNWGLRELEASEPBUFFERDCARBPROC)wglGetProcAddress("wglReleasePbufferDCARB");
	wglDestroyPbufferARB   = (PFNWGLDESTROYPBUFFERARBPROC)wglGetProcAddress("wglDestroyPbufferARB");
	wglQueryPbufferARB     = (PFNWGLQUERYPBUFFERARBPROC)wglGetProcAddress("wglQueryPbufferARB");

	wglBindTexImageARB     = (PFNWGLBINDTEXIMAGEARBPROC)wglGetProcAddress("wglBindTexImageARB");
	wglReleaseTexImageARB  = (PFNWGLRELEASETEXIMAGEARBPROC)wglGetProcAddress("wglReleaseTexImageARB");
	wglSetPbufferAttribARB = (PFNWGLSETPBUFFERATTRIBARBPROC)wglGetProcAddress("wglSetPbufferAttribARB");
	}
	LOG("P Buffer? %s\n", pbuffer?"T":"F");*/

	glGetIntegerv(GL_MAX_LIGHTS, &maxLights);
	LOG("Max Lights? %d\n", maxLights);

	if(strstr(extension, "GL_ARB_multitexture")) {
		glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &maxTextureUnits);

		glMultiTexCoord2fARB = (PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2fARB");
		glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
		glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress("glClientActiveTextureARB");
	}
	LOG("Max Texture Units? %d\n", maxTextureUnits);

	if(strstr(extension, "GL_NV_occlusion_query")) {
		hardwareOcclusion = true;
		glGenOcclusionQueriesNV = (PFNGLGENOCCLUSIONQUERIESNVPROC)wglGetProcAddress("glGenOcclusionQueriesNV");
		glDeleteOcclusionQueriesNV = (PFNGLDELETEOCCLUSIONQUERIESNVPROC)wglGetProcAddress("glDeleteOcclusionQueriesNV");
		glGetOcclusionQueryuivNV = (PFNGLGETOCCLUSIONQUERYUIVNVPROC)wglGetProcAddress("glGetOcclusionQueryuivNV");
		glBeginOcclusionQueryNV = (PFNGLBEGINOCCLUSIONQUERYNVPROC)wglGetProcAddress("glBeginOcclusionQueryNV");
		glEndOcclusionQueryNV = (PFNGLENDOCCLUSIONQUERYNVPROC)wglGetProcAddress("glEndOcclusionQueryNV");
	}
	LOG("Hardware Occlusion? %s\n", hardwareOcclusion?"T":"F");

	if(strstr(extension, "GL_EXT_compiled_vertex_array")) {
		compiledVAs = true;
		glLockArraysEXT = (PFNGLLOCKARRAYSEXTPROC)wglGetProcAddress("glLockArraysEXT");
		glUnlockArraysEXT = (PFNGLUNLOCKARRAYSEXTPROC)wglGetProcAddress("glUnlockArraysEXT");
	}
	LOG("Compiled Vertex Array? %s\n", compiledVAs?"T":"F");

	//if(false) {
	if(strstr(extension, "GL_ARB_vertex_buffer_object")) {
		vertexBufferObjects = true;
		glGenBuffersARB = (PFNGLGENBUFFERSARBPROC) wglGetProcAddress("glGenBuffersARB");
		glBindBufferARB = (PFNGLBINDBUFFERARBPROC) wglGetProcAddress("glBindBufferARB");
		glBufferDataARB = (PFNGLBUFFERDATAARBPROC) wglGetProcAddress("glBufferDataARB");
		glBufferSubDataARB = (PFNGLBUFFERSUBDATAARBPROC) wglGetProcAddress("glBufferSubDataARB");
		glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC) wglGetProcAddress("glDeleteBuffersARB");
	}
	LOG("VBOs? %s\n", vertexBufferObjects?"T":"F");

	//if(false) {
	if(strstr(extension, "GL_EXT_draw_range_elements")) {
		drawRangeElements = true;
		glDrawRangeElementsEXT = (PFNGLDRAWRANGEELEMENTSPROC) wglGetProcAddress("glDrawRangeElementsEXT");
	}
	LOG("Draw Range Elements? %s\n", drawRangeElements?"T":"F");

	if(strstr(extension, "GL_ATI_separate_stencil") && strstr(extension, "GL_EXT_stencil_wrap") ) {
		separateStencilATI = true;
		glStencilOpSeparateATI = (PFNGLSTENCILOPSEPARATEATIPROC) wglGetProcAddress("glStencilOpSeparateATI");
		glStencilFuncSeparateATI = (PFNGLSTENCILFUNCSEPARATEATIPROC) wglGetProcAddress("glStencilFuncSeparateATI");
	}
	LOG("ATI Two Sided Stencil? %s\n", separateStencilATI?"T":"F");

	if(strstr(extension, "GL_EXT_stencil_two_side") && strstr(extension, "GL_EXT_stencil_wrap") ) {
		separateStencil = true;
		glActiveStencilFaceEXT = (PFNGLACTIVESTENCILFACEEXTPROC) wglGetProcAddress("glActiveStencilFaceEXT");
	}
	LOG("Two Sided Stencil? %s\n", separateStencil?"T":"F");

	if(strstr(extension, "GL_EXT_framebuffer_object")) {
		framebufferObject = true;
		glBindRenderbufferEXT = (PFNGLBINDRENDERBUFFEREXTPROC) wglGetProcAddress("glBindRenderbufferEXT");
		glDeleteRenderbuffersEXT = (PFNGLDELETERENDERBUFFERSEXTPROC) wglGetProcAddress("glDeleteRenderbuffersEXT");
		glGenRenderbuffersEXT = (PFNGLGENRENDERBUFFERSEXTPROC) wglGetProcAddress("glGenRenderbuffersEXT");
		glRenderbufferStorageEXT = (PFNGLRENDERBUFFERSTORAGEEXTPROC) wglGetProcAddress("glRenderbufferStorageEXT");
		glBindFramebufferEXT = (PFNGLBINDFRAMEBUFFEREXTPROC) wglGetProcAddress("glBindFramebufferEXT");
		glDeleteFramebuffersEXT = (PFNGLDELETEFRAMEBUFFERSEXTPROC) wglGetProcAddress("glDeleteFramebuffersEXT");
		glGenFramebuffersEXT = (PFNGLGENFRAMEBUFFERSEXTPROC) wglGetProcAddress("glGenFramebuffersEXT");

		glCheckFramebufferStatusEXT = (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC) wglGetProcAddress("glCheckFramebufferStatusEXT");
		glFramebufferTexture2DEXT = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC) wglGetProcAddress("glFramebufferTexture2DEXT");
		glFramebufferRenderbufferEXT = (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC) wglGetProcAddress("glFramebufferRenderbufferEXT");
	}
	LOG("FBOs? %s\n", framebufferObject?"T":"F");

	/*DestroyWindow(hwnd);
	UnregisterClass("Vision3_dummy", hInstance);*/
}
#endif

Shader::~Shader() {
	LOG("deleting %s shader %d\n", vertex_shader?"vertex":"pixel", this);
}

Shader *Shader::createShader(VI_GraphicsEnvironment *genv, const char *file, const char *func,bool vertex_shader) {
	GraphicsEnvironment *d_genv = static_cast<GraphicsEnvironment *>(genv); // cast to internal representation
	return d_genv->createShader(file, func, vertex_shader);
}

void Shader::setDefaults() const {
	// sets some sensible defaults
	const float zero[4] = {0.0f, 0.0f, 0.0f, 0.0f};
	this->setMatSpecular(zero);
}

/*Shader *Shader::createShader(Graphics3D *g3, char *file, char *func,bool vertex_shader) {
	if( !g3->getGraphicsEnvironment()->hasShaders() )
		return NULL;
	Shader *shader = new Shader(g3, vertex_shader);
	shader->cgProgram = cgCreateProgramFromFile(g3->getGraphicsEnvironment()->cgContext, CG_SOURCE, file, shader->cgProfile, func, 0);
	if( shader->cgProgram == NULL ) {
		LOG("### Failed to load vertex shader:\n");
		CGerror Error = cgGetError();
		LOG(cgGetErrorString(Error));
		LOG("\n");
		delete shader;
		g3->getGraphicsEnvironment()->cgContext = NULL;
		return NULL;
	}
	LOG("Successfully loaded %s shader %d\n",vertex_shader?"vertex":"pixel",shader);
	cgGLLoadProgram(shader->cgProgram);
	shader->init();
	g3->addShader(shader);

	return shader;
}*/

//Texture::Texture() : Image2D() {
/*Texture::Texture() {
	//this->filename = NULL;
	//this->mipmap = true;
	//this->w = 0;
	//this->h = 0;
}*/

/*Texture::Texture(bool mipmap) : Image2D() {
this->mipmap = mipmap;
this->textureID = 0;
}*/

Texture::~Texture() {
	//GraphicsEnvironment::checkGLError("Texture::~Texture enter");
	/*if(filename != NULL)
		delete [] filename;*/
	//GraphicsEnvironment::checkGLError("Texture::~Texture exit");
}

FontBuffer::FontBuffer() : VisionObject() {
}

void FontBuffer::writeTextExt(int x,int y,const char *text,...) const {
	va_list vlist;
	const int BUFSIZE = 65535;
	char buff[BUFSIZE+1];
	va_start(vlist, text);
	int length = vsnprintf(buff, BUFSIZE, text, vlist);
	va_end(vlist);
	ASSERT( length >= 0 );
	this->writeText(x,y,buff, length);
}

Graphics2D::Graphics2D(GraphicsEnvironment *genv) : VisionObject(), genv(genv), curr_x(0), curr_y(0), texture(NULL) {
}

Renderer *Graphics2D::getRenderer() const {
	return this->genv->getRenderer();
}

void Graphics2D::move(short x,short y) {
	curr_x = x;
	curr_y = y;
}

//Renderer::Renderer(Graphics3D *g3) : g3(g3), c_pixel_shader(NULL), c_vertex_shader(NULL) {
Renderer::Renderer(GraphicsEnvironment *genv) : genv(genv), static_transformationMatrix(NULL), c_pixel_shader(NULL), c_vertex_shader(NULL), z_near(0.1f), z_far(1000.0f), persp(true) {
//Renderer::Renderer() : c_pixel_shader(NULL), c_vertex_shader(NULL) {
	//this->hdr_fbo = NULL;
	//this->hdr_texture = NULL;
	//this->hdr_depth_stencil_texture = NULL;
	//this->static_transformationMatrix = this->createTransformationMatrix();
	// NB - we create static_transformationMatrix lazily rather than here, as when in a constructor, we can't call the derived classes virtual functions
}

Renderer::~Renderer() {
	delete this->static_transformationMatrix;
}
