#pragma once

#ifdef _WIN32

// disable warning C4250: "xxx inherits yyy via dominance"
#pragma warning(disable:4250)

#define WIN32_MEAN_AND_LEAN
#include <windows.h>

#endif

#ifdef __linux
#include<X11/X.h>
#include<X11/Xlib.h>
#endif

#include "VisionShared.h"
#include "VisionUtils.h"

#include <string>

class VisionPrefs;

class VI_GraphicsEnvironment;
class VI_Graphics2D;
class VI_Font;
class VI_Image;
class VI_Texture;

class VI_World;
class VI_Object;
class VI_VObject;
class VI_SceneGraphNode;
class VI_Entity;
class VI_Face;
class VI_Terrain;
class VI_Sky;

class VI_Panel;

class Renderer;

/*VI_GraphicsEnvironment *V_GRAPHICSENV_get_ptr(V_TAG_t v_genv);
VI_Panel *V_PANEL_get_ptr(V_TAG_t v_panel);
VI_World *V_WORLD_get_ptr(V_TAG_t v_world);
VI_Sky *V_SKY_get_ptr(V_TAG_t v_sky);
VI_Font *V_FONT_get_ptr(V_TAG_t v_font);
VI_Image *V_IMAGE_get_ptr(V_TAG_t v_image);
VI_Entity *V_ENTITY_get_ptr(V_TAG_t v_entity);
VI_Object *V_OBJECTDATA_get_ptr(V_TAG_t v_object);
VI_VObject *V_VOBJECT_get_ptr(V_TAG_t v_vobject);
VI_Terrain *V_TERRAIN_get_ptr(V_TAG_t v_terrain);
VI_Graphics2D *V_GRAPHICS2D_get_ptr(V_TAG_t v_g2d);*/

void VI_addtag();

void VI_initialise(const char *application_name);
char *VI_getApplicationFilename(const char *name);
void VI_cleanup();
void VI_check();
void VI_set_persistence_all(int level);
void VI_set_default_persistence(int level);
void VI_debug_mem_usage_all();
void VI_lock();
void VI_unlock();
bool VI_islocked();
void VI_flush(int level);
void VI_flush_all();
void VI_update(int time);
bool VI_keydown(int v_keycode);
bool VI_keyany();
bool VI_keypressed(int v_keycode);
void VI_createThreads(int n_threads, VI_ThreadFunction **function, void **data);

#ifndef VI_log
#define VI_log V_log
#endif

class VI_Sound {
public:
	virtual ~VI_Sound() {}

	virtual bool LoadSound(int s,const char *file)=0;
	virtual bool PlaySound(int s,bool loop,int delay)=0;
	virtual void FreeSound(int s)=0;
	virtual bool IsSoundPlaying(int s)=0;
	virtual void StopSound(int s,bool bResetPosition)=0;

	virtual void playMedia(const char *filename)=0;
	virtual bool isMediaPlaying()=0;

#ifdef _WIN32
	virtual void processEvents(UINT message, WPARAM wParam, LPARAM lParam)=0;
	virtual bool initHWND(HWND hWnd)=0;
	//static VI_Sound *create(HWND hwnd);
#endif
	static VI_Sound *create();

};

class VI_Graphics2D {
public:
	virtual ~VI_Graphics2D() {}

	virtual void setColor3i(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setColor4i(unsigned char r, unsigned char g, unsigned char b, unsigned char a)=0;
	virtual void setTexture(VI_Texture *texture)=0;
	virtual void move(short x,short y)=0;
	virtual void plot(short x,short y)=0;
	virtual void draw(short x,short y)=0;
	virtual void draw(short x1,short y1,short x2,short y2) const=0;
	virtual void drawRect(short x,short y,short width,short height) const=0;
	virtual void fillRect(short x,short y,short width,short height) const=0;
};

typedef bool (VI_GraphicsEnvironment_Update) (int time);
typedef void (VI_GraphicsEnvironment_Render) ();

class VI_GraphicsEnvironment {
public:
	virtual ~VI_GraphicsEnvironment() {}
	virtual V_TAG_t getTag() const=0;

	virtual VI_Font *createFont(const char *family,int pt,int style) const=0;

	virtual Renderer *getRenderer() const=0;
	virtual int getWidth() const=0;
	virtual int getHeight() const=0;
	virtual int getAA() const=0;
	virtual bool isFullScreen() const=0;
	virtual const VisionPrefs *getVisionPrefs() const=0;
	virtual void setSMP(bool smp)=0;
	virtual bool hasShaders() const=0;
	virtual bool isActive() const=0;
	virtual void getFrameStats(int *poly_count,int *vertex_count, int *distinct_vertex_count, int *render_calls_count) const=0;
	virtual float getFPS() const=0;
	virtual void getMouseInfo(int *x,int *y,bool buttons[3]) const=0;
	virtual void getMouseClick(bool buttons[3]) const=0;
	virtual bool readMouseWheelUp() const=0;
	virtual bool readMouseWheelDown() const=0;

	virtual VI_Panel *getPanel() const=0;

	virtual void setZ(float z_near,float z_far)=0;
	virtual void setMouseTexture(VI_Texture *mouse_texture)=0;
	virtual void setPanel(VI_Panel *panel)=0;
	virtual void setModalPanel(VI_Panel *panel)=0;
	virtual VI_Panel *getModalPanel() const=0;
	virtual void setUseShadowVolumes(bool use_shadow_volumes)=0;
	virtual void linkWorld(VI_World *world)=0;
	virtual VI_World *getWorld()=0;
	virtual const VI_World *getWorld() const=0;
	virtual void setSound(VI_Sound *sound)=0;
	virtual VI_Sound *getSound() const=0;

	virtual Vector3D getDirectionOfPoint(VI_SceneGraphNode *node, int sx, int sy) const=0;
	virtual void localToScreen(int *x, int *y, VI_SceneGraphNode *node, const Vector3D *local) const=0;
	virtual Vector3D worldToEye(const Vector3D *world) const=0;
	virtual void worldToScreen(int *x,int *y,const Vector3D *world) const=0;

	virtual void setVisible(bool visible)=0;
	virtual void setWindowTitle(const char *title)=0;
	virtual void showCursor(bool show)=0;
	virtual void setCursor(unsigned char *data, unsigned char *mask, int w, int h, int hot_x, int hot_y)=0;

	virtual void drawFrame()=0;
	virtual void drawFrameTest()=0; // debugging use only
	virtual void swapBuffers()=0;
	virtual void handleInput()=0;
	virtual void handleInput(int mouseX,int mouseY,bool mouseL,bool mouseM,bool mouseR)=0;
	virtual void flushInput()=0;
	virtual void processEvents()=0;
	virtual void waitForEvents()=0;
	virtual void sleep(int time)=0; // sleep, time in milliseconds
	virtual bool shouldQuit()=0;
	virtual void clearQuitRequest()=0;
	virtual bool saveScreen(const char *filename,const char *type)=0;
	virtual void render(VI_Entity *entity, int current_time, float x, float y, float z)=0;
	virtual void setRenderFunc(VI_GraphicsEnvironment_Render *renderFunc)=0;
	virtual void setFade(float duration, unsigned char fade_start, unsigned char fade_end)=0;
	virtual void stopFade()=0;

	void render(VI_GraphicsEnvironment_Update update);

	// OS specific functions
#ifdef _WIN32
	virtual HWND getHWND() const=0;
	virtual bool isTrueFullscreen() const=0; // hack for Windows, whether the full screen has been implemented as a "true" fullscreen mode in Windows
#endif
#ifdef __linux
	virtual Display *getXDisplay() const=0;
	virtual Window getXWindow() const=0;
#endif
};

class VI_Texture {
public:
	virtual ~VI_Texture() {}
	virtual V_TAG_t getTag() const=0;

	virtual int getWidth() const=0;
	virtual int getHeight() const=0;

	virtual void draw(int x, int y)=0;
	virtual void draw(int x, int y, int w, int h)=0;
	virtual void enable()=0;
};

class VI_Font {
public:
	virtual ~VI_Font() {}
	virtual V_TAG_t getTag() const=0;

	virtual void writeText(int x,int y,const char *text) const=0;
	virtual void writeText(int x,int y,const char *text,int length) const=0;
	virtual void writeTextExt(int x,int y,const char *text,...) const=0;
	virtual int getWidth(char letter) const=0;
	virtual int getWidth(const char *text) const=0;
	virtual int getHeight() const=0;
};

class VI_Image {
public:
	virtual ~VI_Image() {}
	virtual V_TAG_t getTag() const=0;

	virtual int getWidth() const=0;
	virtual int getHeight() const=0;
	virtual bool getAlpha() const=0;
	virtual bool makePowerOf2()=0;
	virtual bool makeSamePowerOf2()=0;
	virtual unsigned char *getData()=0;
	virtual void putRGB(int x, int y, unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void putRGBA(int x, int y, unsigned char r, unsigned char g, unsigned char b, unsigned char a)=0;
	virtual void getRGB(unsigned char *r, unsigned char *g, unsigned char *b, int x, int y) const=0;
	virtual void getRGBA(unsigned char *r, unsigned char *g, unsigned char *b, unsigned char *a, int x, int y) const=0;
	virtual const void *get(int x,int y) const=0;
	virtual const unsigned char *getData() const=0;

	virtual void multiply(float scaleR, float scaleG, float scaleB)=0;
	virtual void adjustBrightness(unsigned char bright)=0;
	virtual void merge(VI_Image *image)=0;
	virtual bool scale(int newWidth, int newHeight)=0;
	virtual void flip(bool flipX, bool flipY)=0;

	virtual VI_Image *copy() const=0;
	virtual VI_Image *createDerivative(float scale) const=0;
	virtual VI_Image *createReflection2D() const=0;
	virtual bool save(const char *filename,const char *ext) const=0;

	static VI_Image *createNoise(int w,int h);
	static VI_Image *createNoise(int w,int h,float scale_u,float scale_v,const unsigned char filter_max[3],const unsigned char filter_min[3],V_NOISEMODE_t noisemode,int n_iterations);
	static VI_Image *createRadial(int w,int h,const unsigned char filter_max[3],const unsigned char filter_min[3],bool squared);
};

class VI_Sky {
public:
	virtual ~VI_Sky() {}
	virtual V_TAG_t getTag() const=0;

	virtual void setMode(V_SKYMODE_t mode)=0;
	virtual void setColori(unsigned char r,unsigned char g,unsigned char b)=0;

	virtual void render(Renderer *renderer)=0;
};

class VI_Clouds {
public:
	virtual ~VI_Clouds() {}
	virtual V_TAG_t getTag() const=0;

	virtual void setColori(unsigned char r,unsigned char g,unsigned char b)=0;
};

class VI_Atmosphere {
public:
	virtual ~VI_Atmosphere() {}
	virtual V_TAG_t getTag() const=0;
};

class VI_Terrain {
public:
	virtual ~VI_Terrain() {}
	virtual V_TAG_t getTag() const=0;

	virtual int getWidth() const=0;
	virtual int getDepth() const=0;
	virtual float getXScale() const=0;
	virtual float getZScale() const=0;
	virtual float getHScale() const=0;
	virtual unsigned char get(int x,int z) const=0;
	virtual void set(int x,int z,unsigned char v)=0;
	virtual void setTextureMap(int x,int z,unsigned char v)=0;
	virtual float getHeightAt(float x,float z) const=0;
	virtual Vector3D getNormalAt(float x,float z) const=0;
	virtual bool rayfire(Vector3D *hit, Vector3D start, Vector3D dir, float range) const=0;

	virtual void setOffset(Vector3D offset)=0;
	virtual void setLOD(bool lod)=0;
	virtual void setTexture(int index,VI_Texture *texture)=0;
	virtual void setSea(bool sea,float sea_height,float sea_extent,bool sea_refl)=0;
	virtual void setSeaTexture(VI_Texture *sea_texture)=0;
	virtual void setLighting(bool lighting_enabled)=0;

	virtual void setColorMode(V_COLORMODE_t colormode)=0;
	//virtual void setColorNoise(int index,float min_r,float min_g,float min_b,float max_r,float max_g,float max_b,float noise_scale)=0;
	virtual void setColorNoisei(int index,unsigned char min_r,unsigned char min_g,unsigned char min_b,unsigned char max_r,unsigned char max_g,unsigned char max_b,float noise_scale)=0;
	virtual void setColorAmbientOcclusion()=0;

	virtual bool isVisible(int x,int z) const=0;
	virtual void setVisible(int x,int z,bool v)=0;

	virtual int getNodesRendered() const=0;
};

class VI_World {
public:
	virtual ~VI_World() {}
	virtual V_TAG_t getTag() const=0;

	virtual void getAmbience(float *r,float *g,float *b)=0;
	virtual VI_SceneGraphNode *getViewpoint()=0;
	virtual const VI_SceneGraphNode *getViewpoint() const=0;

	virtual bool addNode(VI_SceneGraphNode *node)=0;
	virtual void setSky(VI_Sky *sky)=0;
	virtual VI_Sky *getSky()=0;
	virtual const VI_Sky *getSky() const=0;
	virtual void setClouds(VI_Clouds *clouds)=0;
	virtual VI_Clouds *getClouds()=0;
	virtual const VI_Clouds *getClouds() const=0;
	virtual void setAtmosphere(VI_Atmosphere *atmosphere)=0;
	virtual VI_Atmosphere *getAtmosphere()=0;
	virtual const VI_Atmosphere *getAtmosphere() const=0;
	//virtual void setHorizon(bool enabled)=0;
	virtual void setBackColori(unsigned char r, unsigned char g, unsigned char b)=0;
	//virtual void setGroundColori(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setTerrain(VI_Terrain *terrain)=0;
	virtual VI_Terrain *getTerrain()=0;
	virtual const VI_Terrain *getTerrain() const=0;
	virtual void setAmbiencei(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setViewpoint(VI_SceneGraphNode *viewpoint)=0;
	virtual void enableFog(bool enabled)=0;
	virtual void setFogParmsExp(float density)=0;
/*
V_STATUS_t V_WORLD_add_entity(V_WORLD_t world,V_ENTITY_t entity);
V_STATUS_t V_WORLD_add_static_geometry(V_WORLD_t v_world, const V_OBJECTDATA_t *v_objects, int n_objects);
V_STATUS_t V_WORLD_remove_all(V_WORLD_t world);
V_STATUS_t V_WORLD_set_terrain(V_WORLD_t v_world,V_TERRAIN_t v_terrain);
V_STATUS_t V_WORLD_enable_fog(V_WORLD_t world,bool fog);
V_STATUS_t V_WORLD_set_fog_color(V_WORLD_t world,float r,float g,float b);
V_STATUS_t V_WORLD_set_fog_parms_exp(V_WORLD_t world,float density);
V_STATUS_t V_WORLD_get_viewpoint(V_WORLD_t v_world,V_ENTITY_t *v_entity);
V_STATUS_t V_WORLD_set_viewpoint(V_WORLD_t v_world,V_ENTITY_t v_entity);
*/
};

class VI_HAnimationInfo {
public:
	virtual ~VI_HAnimationInfo() {}
	// note, no tag!

	virtual void setName(const char *name)=0;
	virtual const char *getName() const=0;
	virtual void addPositionKey(float time,Vector3D pos)=0;
	virtual void addRotationKey(float time,Rotation3D rot)=0;

	static VI_HAnimationInfo *create(const char *name, float start, float end);
};

class VI_SceneGraphNode {
public:
	virtual ~VI_SceneGraphNode() {}
	virtual V_TAG_t getTag() const=0;
	virtual void setName(const char *name)=0;
	virtual const char *getName() const=0;

	virtual VI_SceneGraphNode *copy() const=0;
	virtual int getNChildren() const=0;
	virtual VI_SceneGraphNode *getChild(int i) const=0;
	virtual VI_SceneGraphNode *getParent() const=0;
	virtual VI_Object *getObject() const=0;
	virtual Vector3D getPosition() const=0;
	virtual Vector3D getCurrentWorldPos() const=0;
	virtual Vector3D getLastRenderedEyePos() const=0;
	virtual Vector3D localToWorldPos(const Vector3D *pos) const=0;
	virtual Vector3D worldToLocalDir(const Vector3D *dir) const=0;
	virtual Rotation3D getRotation() const=0;
	virtual bool getExtent(float *low,float *high,const Vector3D *p,const Vector3D *n,bool firstFrameOnly,bool do_children) const=0;
	virtual bool isVisible() const=0;
	virtual V_SOURCETYPE_t getSourcetype() const=0;
	virtual void setSourcetype(V_SOURCETYPE_t sourcetype)=0;

	virtual void setPosition(float x,float y,float z)=0;
	virtual void setPosition(const Vector3D *v)=0;
	virtual void setInfinite(bool infinite)=0;
	virtual void move(float x,float y,float z)=0;
	virtual void move(Vector3D *v)=0;
	virtual void moveDirection(Vector3D v,float d)=0;
	virtual void moveDirectionY(Vector3D v,float d)=0;
	virtual void rotate(float x,float y,float z)=0;
	virtual void rotateLocal(float x,float y,float z)=0;
	virtual void rotateToEuler(float x,float y,float z)=0;
	virtual void setFrameSpeed(float frame_speed)=0; // in frames per second
	virtual void setVisible(bool visible)=0;
	virtual void addNode(VI_SceneGraphNode *node)=0;
	virtual VI_SceneGraphNode *findNode(const char *name)=0;
	virtual void kill()=0;
	virtual void setSpeculari(unsigned char r, unsigned char g, unsigned char b, bool do_children)=0;
	virtual void setShadowCaster(bool enabled,bool do_children)=0;
	virtual void enableParticlesystem(bool enabled,bool do_children)=0;
	virtual void expireWhenPsDone(bool enabled)=0;
	virtual void removeChildren() const=0;
	virtual void inheritHAnimationFrom(VI_SceneGraphNode *from)=0;
	virtual bool setHAnimation(const char *animname, bool forcereset)=0;
	virtual void setHAnimationSpeed(float animationspeed)=0; // in frames per second
	virtual void addHAnimationInfo(VI_HAnimationInfo *animationinfo)=0;
	virtual float getHAnimationTime() const=0;
	virtual void setHAnimationTime(float animationtime, bool do_children)=0;
	virtual void adoptHAnimation(bool do_children)=0;
	virtual void setDefaultHAnimationPosition(bool do_children)=0;

	virtual bool isAnimationFinished() const=0;
	virtual void setAnimLoop(bool loop)=0;

	virtual void setAllToIdentity(bool do_this)=0;
	virtual void setVelocity(float x,float y,float z)=0;
	virtual void setForce(float x,float y,float z)=0;
	/*virtual void getVelocity(float *x,float *y,float *z) const=0;
	virtual void getForce(float *x,float *y,float *z) const=0;*/
	virtual Vector3D getVelocity() const=0;
	virtual Vector3D getForce() const=0;
	virtual void halt()=0;

	virtual bool collision(const VI_SceneGraphNode *node) const=0;
	virtual bool collisionLine(float *intersection, const Vector3D *pos, const Vector3D *dir) const=0;

	virtual void update(int time, bool do_children)=0;
};

//typedef void (VI_Entity_Update) (VI_Entity *source, float time);
typedef void (VI_Entity_Update) (VI_Entity *source, int time);

class VI_Entity : public virtual VI_SceneGraphNode {
public:
	virtual ~VI_Entity() {}

	virtual int getState() const=0;
	//virtual void getForceColor(V_BLENDTYPE_t *blendtype,float *r,float *g,float *b,float *alpha) const=0;
	virtual void getForceColori(V_BLENDTYPE_t *blendtype,unsigned char *r,unsigned char *g,unsigned char *b,unsigned char *alpha) const=0;
	virtual void *getUpdateData()=0;

	virtual void setState(int state)=0;
	virtual void setTextureSpeed(float texture_x_sp, float texture_y_sp)=0; // in frames per second
	virtual void forceColori(unsigned char r, unsigned char g, unsigned char b, unsigned char alpha)=0;
	virtual void unforce()=0;
	virtual void setBlendtype(V_BLENDTYPE_t blendtype)=0;
	virtual void setUpdateFunc(VI_Entity_Update *updateFunc)=0;
};

class VI_ShadowPlane {
public:
	virtual ~VI_ShadowPlane() {}
	virtual V_TAG_t getTag() const=0;
};

class VI_Face {
public:
	enum LightingStyle {
		FLAT = 0,
		CURVED = 1
	};

	virtual ~VI_Face() {}
	// note, no tag!

	virtual void setName(const char *name)=0;
	virtual const char *getName() const=0;

	virtual int getNVertices() const=0;
	virtual int getVertexIndex(int i) const=0;

	virtual void setTexture(VI_Texture *texture)=0;
};

/*class VI_Vertex3D {
public:
	virtual ~VI_Vertex3D() {}
	// note, no tag!
};*/

typedef Vector3D (VI_Deform) (Vector3D,void *data);

class VI_Object {
public:
	virtual ~VI_Object() {}
	virtual V_TAG_t getTag() const=0;
	virtual void setName(const char *name)=0;
	virtual bool check() const=0;

	virtual Sphere getBoundingSphere()=0;
	virtual Box getBoundingAABB()=0;
	virtual void getExtent(float *low,float *high,const Vector3D *n,bool firstFrameOnly) const=0;
	virtual void getDirectionRadiusSq(float *radius_sq,Vector3D *n,bool firstFrameOnly) const=0;
	virtual Vector3D findCentre() const=0;
	virtual bool isShadowCaster() const=0;

	virtual void scale(float sx,float sy,float sz)=0;
	virtual void translate(float sx,float sy,float sz)=0;
	virtual void rotate(Rotation3D *rot)=0;
	virtual void rotateEuler(float x,float y,float z)=0;
	virtual void transform4d(Matrix4d *m)=0;
	virtual void deform(VI_Deform *deformFunc, void *deformData)=0;
	virtual void setSpeculari(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setShadowPlane(VI_GraphicsEnvironment *genv, VI_ShadowPlane *shadowPlane, unsigned char alpha)=0;
	virtual void setShadowCaster(bool shadow_caster)=0;
	virtual int getNFrames() const=0;
	virtual void setAnimFrames(V_ANIMSTATE_t state,int start,int end)=0;
	virtual void getFrames(V_ANIMSTATE_t state,int *start,int *end) const=0;
};

class TextureUV {
public:
	//int vx_indx;
	float u;
	float v;
	TextureUV(/*int vx_indx,*/float u, float v) {
		//this->vx_indx = vx_indx;
		this->u = u;
		this->v = v;
	}
};

class VI_Mesh : public virtual VI_Object {
public:
	virtual ~VI_Mesh() {}

	virtual void addTexture(VI_Texture *texture)=0;
	virtual VI_Mesh *clone() const=0;
	virtual void combineObject(VI_Mesh *mesh)=0;

	static VI_Mesh *create(const char *name);
	static VI_Mesh *create(const char *name, int n_frames);
};

class VI_VObject : public virtual VI_Object {
public:
	virtual ~VI_VObject() {}

	virtual int get_n_vertices() const=0;
	virtual int get_n_polys() const=0;

	virtual int addVertex(float x, float y, float z)=0;
	virtual int addVertex(int frame, float x, float y, float z)=0;
	virtual void setVertexColori(int index, unsigned char r, unsigned char g, unsigned char b)=0;

	virtual VI_Face *addPolygon(int v0, int v1, int v2)=0;
	virtual VI_Face *addPolygon(int v0, int v1, int v2, int v3)=0;
	virtual VI_Face *addPolygon(int v0, int v1, int v2, const char *name)=0;
	virtual VI_Face *addPolygon(int v0, int v1, int v2, int v3, const char *name)=0;
	virtual VI_Face *addPolygon(const int *vlist, int n_vertices, const char *name)=0;
	virtual void reservePolygons(int n_polys)=0;
	virtual void clearPolygons()=0;

	virtual void setTexture(VI_Texture *texture) {
		this->setTexture(texture, false, 0, false, 0.0f, 0.0f);
	}
	virtual void setTexture(VI_Texture *texture, float size_u, float size_v) {
		this->setTexture(texture, false, 0, true, size_u, size_v);
	}
	virtual void setTexture(VI_Texture *texture,bool mirror,int skip,bool have_size,float size_u,float size_v)=0;
	virtual void setTextureWrap(VI_Texture *texture, TextureUV **text_uvs)=0;
	virtual void setPolygonTextureWrap(VI_Face *face,VI_Texture *texture,TextureUV **text_uvs)=0;
	virtual void setPolygonTextureCoords(VI_Face *face,VI_Texture *texture,const float *u, const float *v)=0;
	virtual void setBumpTexture(const VI_GraphicsEnvironment *genv, VI_Texture *bump_texture)=0;

	virtual void combineObject(VI_VObject *o)=0;
	virtual VI_VObject *clone() const=0;

	virtual Vector3D getVertexPos(int index) const=0;
	virtual int getFaces(VI_Face ***faces)=0;
	virtual const VI_Face *getFace(int index) const=0;
	virtual VI_Face *getFace(int index)=0;
	virtual VI_Face *findFace(const char *name)=0;
	virtual void setShading(VI_Face::LightingStyle shading)=0;
	virtual void setBaseColori(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setPolygonBaseColori(VI_Face *face, unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setPolygonTexture(VI_Face *face,VI_Texture *texture,bool mirror,int skip,bool have_size,float size_u,float size_v)=0;
	virtual void setPolygonTexture(VI_Face *face,VI_Texture *texture,const Vector3D *origin, const Vector3D *u_dir, const Vector3D *v_dir)=0;
	virtual void invertPolygon(VI_Face *face)=0;
	virtual void setSolid(bool solid)=0;
	virtual bool isSolid() const=0;

	virtual void subdivide()=0;

	//static VI_VObject *create();
	static VI_VObject *create(const char *name);
	static VI_VObject *create(const char *name, int n_frames);
};

class VI_Billboard : public virtual VI_Object {
public:
	virtual ~VI_Billboard() {}

	virtual void setOffset(Vector3D v, int i)=0;
	virtual void setLock(bool lock)=0;
	virtual void setWave(float wave_amp, float wave_sp)=0;
	virtual void setColori(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setOneColori(int i, unsigned char r, unsigned char g, unsigned char b)=0;
};

class VI_Light : public virtual VI_Object {
public:
	virtual ~VI_Light() {}

	virtual void setLightAmbienti(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setLightDiffusei(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setLightSpeculari(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setLightAttenuation(float quadratic,float linear,float constant)=0;
};

class VI_Explosion : public virtual VI_Light {
public:
	virtual ~VI_Explosion() {}
};

class VI_Particlesystem : public virtual VI_Object {
public:
	virtual ~VI_Particlesystem() {}

	virtual void setTexture(VI_Texture *texture)=0;

	static VI_Particlesystem *createSnowstorm(int max_particles,VI_Texture *texture,float width,float height,float depth);
};

class VI_GeneralParticlesystem : public virtual VI_Particlesystem {
public:
	virtual ~VI_GeneralParticlesystem() {}

	virtual void setColoriStart(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setColoriEnd(unsigned char r, unsigned char g, unsigned char b)=0;
	virtual void setAlphaiStart(unsigned char a)=0;
	virtual void setAlphaiEnd(unsigned char a)=0;
	virtual void setSizeStart(float size)=0;
	virtual void setSizeEnd(float size)=0;
	virtual void setBirthRate(int value)=0;
	virtual void setLife(float value)=0;
	virtual void setMass(float value)=0;
	virtual void setInitialVelocity(float value)=0;

	static VI_GeneralParticlesystem *create(int max_particles);
};

/*class VI_Modeller {
public:
	static VI_VObject *createGrid(float w,float d,unsigned short nx,unsigned short ny,bool mirror,bool share);
	static VI_VObject *createCuboid(float w,float h,float d);
	static VI_VObject *createSphere(float r,unsigned short d);

	static void deformAttractor(VI_Object *obj, Vector3D centre, float radius, float scale);
};*/

typedef void (VI_Panel_Paint) (VI_Graphics2D *g2d, void *data);
typedef void (VI_Panel_Action) (VI_Panel *source);
//typedef void (VI_Panel_Action) (VI_Panel *source, void *data);

class VI_Panel {
public:
	virtual ~VI_Panel() {}
	virtual V_TAG_t getTag() const=0;
	virtual void setName(const char *name)=0;
	virtual const char *getName() const=0;

	virtual void setEnabled(bool enabled,bool do_children)=0;
	virtual bool isEnabled() const=0;
	virtual void setVisible(bool visible)=0;
	virtual bool isVisible() const=0;
	virtual void setHasInputFocus(bool has_input_focus)=0;

	virtual void setPopupText(const char *popup_text, const VI_Font *font)=0;
	virtual void add(VI_Panel *panel,int x,int y)=0;
	virtual void remove(VI_Panel *panel)=0;
	virtual void removeAll()=0;
	virtual void detach()=0;
	//virtual void setAction(V_Action *action)=0;
	virtual void setMeasureFrom(bool from_left, bool from_top)=0;
	virtual void setAction(VI_Panel_Action *action)=0;
	virtual void setMouseEnterAction(VI_Panel_Action *mouse_enter_action)=0;
	virtual void setMouseExitAction(VI_Panel_Action *mouse_exit_action)=0;
	virtual void setPaintFunc(VI_Panel_Paint *paintPanel,void *paintData)=0;
	virtual void setKeyShortcut(int shortcut)=0;
	virtual void disableKeyShortcut()=0;
	virtual void setForeground(float r,float g,float b)=0;
	virtual void setForeground(float r,float g,float b,float a)=0;
	virtual void setBackground(float r,float g,float b)=0;
	virtual void setBackground(float r,float g,float b,float a)=0;
	virtual void setBackgroundAlphai(int alpha)=0;
	virtual void setTexture(VI_Texture *texture)=0;
	virtual void setSize(int width, int height)=0;
	virtual void getSize(int *width, int *height) const=0;
	virtual int getWidth() const=0;
	virtual int getHeight() const=0;
	virtual void setUserData(void *userData)=0;
	virtual void *getUserData() const=0;
	virtual VI_Panel *find(int fx, int fy)=0;
};

class VI_Button : public virtual VI_Panel {
public:
	virtual ~VI_Button() {}

	virtual void setBorder(bool border)=0;
};

class VI_ImageButton : public virtual VI_Panel {
public:
	virtual ~VI_ImageButton() {}
};

class VI_Cycle : public virtual VI_Panel {
public:
	virtual ~VI_Cycle() {}

	virtual int getActive() const=0;
	virtual int getNEntries() const=0;
	virtual void addEntry(const char *entry)=0;

	virtual void setActive(int active)=0;
};

class VI_Listbox : public virtual VI_Panel {
public:
	virtual ~VI_Listbox() {}

	virtual int getNEntries() const=0;
	virtual int getActive() const=0;
	virtual void setActive(int active)=0;
	virtual void reset(const std::string *entries, int n_entries)=0;
	virtual void reset(char **entries)=0;
};

class VI_Stringgadget : public virtual VI_Panel {
public:
	virtual ~VI_Stringgadget() {}

	virtual void setReadOnly(bool read_only)=0;
	virtual void setMaxLen(int max_len)=0;
	virtual void addText(const char *text)=0;
	virtual void clear()=0;
	virtual void getText(char **text,int *length) const=0;
	virtual const char *getText() const=0;
	virtual void setEnterAction(VI_Panel_Action *action)=0;
};

class VI_Textfield : public virtual VI_Stringgadget {
public:
	virtual ~VI_Textfield() {}

	virtual void setCPos(int cpos)=0;
};

VI_GraphicsEnvironment *VI_createGraphicsEnvironment(int width,int height,int multisample,bool fullscreen,bool hdr);
VI_GraphicsEnvironment *VI_createGraphicsEnvironment(VisionPrefs *visionPrefs);

VI_World *VI_createWorld();

VI_Sky *VI_createSky(const char *filename,const char *ext);
VI_Sky *VI_createSky1(const char *filename);
VI_Sky *VI_createSky1(VI_Texture *texture);

VI_Clouds *VI_createClouds();

VI_Atmosphere *VI_createAtmosphere(VI_Entity *light_entity);

VI_Terrain *VI_createTerrain(VI_GraphicsEnvironment *genv,int width,int depth,float xscale,float zscale,float hscale);
VI_Terrain *VI_createTerrainRandom(VI_GraphicsEnvironment *genv,int width,int depth,float xscale,float zscale,float hscale,const int base_heights[4],int min_height,int max_height,float roughness,int seed);

VI_Image *VI_createImage(const char *filename);
VI_Image *VI_createImage(int w,int h);
VI_Image *VI_createImageNoise(int w,int h);
VI_Image *VI_createImageNoise(int w,int h,float scale,const unsigned char filter_max[3],const unsigned char filter_min[3],V_NOISEMODE_t noisemode,int n_iterations);
VI_Image *VI_createImageNoise(int w,int h,float scale_u,float scale_v,const unsigned char filter_max[3],const unsigned char filter_min[3],V_NOISEMODE_t noisemode,int n_iterations);
VI_Image *VI_createImageRadial(int w,int h,const unsigned char filter_max[3],const unsigned char filter_min[3],bool squared);

VI_Texture *VI_createTexture(const char *filename);
VI_Texture *VI_createTextureWithMask(const char *filename,unsigned char r,unsigned char g,unsigned char b);
VI_Texture *VI_createTexture(VI_Image *image);
VI_Texture *VI_createTextureWithMask(VI_Image *image,unsigned char r,unsigned char g,unsigned char b);

/*VI_VObject *VI_createCuboid(float w,float h,float d);
VI_VObject *VI_createGrid(float w,float d,unsigned short nx,unsigned short ny,bool mirror,bool share);*/

VI_Billboard *VI_loadBillboardWithMask(const char *filename,float width,float height,unsigned char r,unsigned char g,unsigned char b,int n_billboards);

VI_Entity *VI_createLight(VI_Light **light,float aR,float aG,float aB,float dR,float dG,float dB,float x,float y,float z);

VI_Entity *VI_createExplosion(VI_Explosion **explosion,VI_Texture *texture,float aR,float aG,float aB,float dR,float dG,float dB,float x,float y,float z);

//VI_Entity *VI_createEntity();
VI_SceneGraphNode *VI_createSceneGraphNode();
VI_Entity *VI_createEntity(VI_Object *obj);

VI_ShadowPlane *VI_createShadowPlane(Vector3D normal,float D,bool shadow,bool reflection);

VI_Panel *VI_createPanel();
VI_Button *VI_createButton(const char *text,const VI_Font *font);
VI_ImageButton *VI_createImageButton(VI_Texture *texture);
VI_ImageButton *VI_createImageButton(VI_Texture *texture, int width, int height);
VI_Cycle *VI_createCycle(char **entries,const VI_Font *font);
VI_Cycle *VI_createCycle(const VI_Font *font);
VI_Listbox *VI_createListbox(int width,int height,char **entries,const VI_Font *font);
VI_Listbox *VI_createListbox(int width,int height,const std::string *entries,int n_entries,const VI_Font *font);
VI_Stringgadget *VI_createStringgadget(const VI_Font *font,int width);
VI_Textfield *VI_createTextfield(const VI_Font *font,int width,int height);

void VI_movementSimpleXY(VI_SceneGraphNode *node,int time);
void VI_movementSimpleXYhover(VI_SceneGraphNode *node,int time);

//#include "VisionTest.h"
