//---------------------------------------------------------------------------
#include <ctime>

#include "World.h"
#include "Entity.h"
#include "Terrain.h"

//---------------------------------------------------------------------------

World::World() : VisionObject() {
	genv = NULL;
	//groundLevel = 0.3f;

	//bsptree = NULL;
	terrain = NULL;
	skybox = NULL;
	clouds = NULL;
	atmosphere = NULL;
	sector.rootObject = new SceneGraphNode();
	sector.rootObject->setWorld(this);
	lightAmbient.seti(204,204,204);
	backColor.seti(0,0,0);
	//groundColor.seti(0,255,0);
	//enableHorizon = false;
	fog.enable(false);

	defViewPoint = new SceneGraphNode();
	viewPoint = defViewPoint;
	sector.rootObject->addNode(viewPoint);

	//sector.spatialindex = new SpatialGrid2D(-5.0, -10.0, -5.0, 330.0, 20.0, 330.0, 8, 8);
	//sector.spatialindex = new Quadtree(-5.0, -10.0, -5.0, 330.0, 20.0, 330.0, 5);

	timer = clock();
	//Vision::addWorld(this);
}

World::~World() {
	//int i=0;
	/*for(i=0;i<n_obj;i++)
	delete rootObjects[i];*/
	if( sector.rootObject != NULL )
		delete sector.rootObject;
	if( sector.spatialindex != NULL )
		delete sector.spatialindex;
	/*delete shadowNodes;
	delete updateNodes;*/
	//for(i=0;i<n_tobj;i++)
	//	delete tempObjects[i];
	//for(i=0;i<n_tex;i++)
	//	delete textures[i];
	//for(i=0;i<n_lights;i++)
	//	delete lights[i];
	//delete movement;
	/*if(defViewPoint)
	delete defViewPoint;*/
	//Vision::removeWorld(this);
}

int World::memUsage() {
	int size = sizeof(this);
	/*size += shadowNodes->size() * sizeof(void *);
	size += updateNodes->size() * sizeof(void *);*/
	return size;
}


void World::setGraphicsEnvironment(GraphicsEnvironment *genv) {
	this->genv=genv;
	//this->inputHandler->setGraphicsEnvironment(genv);
}

bool World::addNode(VI_SceneGraphNode *node) {
	if(getRootNode() != NULL) {
		//SceneGraphNode *d_node = static_cast<SceneGraphNode *>(node); // cast to the internal representation
		SceneGraphNode *d_node = dynamic_cast<SceneGraphNode *>(node); // cast to the internal representation
		getRootNode()->addNode(d_node);
		if( this->sector.spatialindex != NULL ) {
			this->sector.spatialindex->addNode(d_node);
		}
		return true;
	}
	return false;
}

void World::setSky(VI_Sky *sky) {
	SkyBox *d_sky = static_cast<SkyBox *>(sky); // cast to the internal representation
	this->skybox = d_sky;
}

void World::setClouds(VI_Clouds *clouds) {
	Clouds *d_clouds = static_cast<Clouds *>(clouds); // cast to the internal representation
	this->clouds = d_clouds;
}

void World::setAtmosphere(VI_Atmosphere *atmosphere) {
	Atmosphere *d_atmosphere = static_cast<Atmosphere *>(atmosphere); // cast to the internal representation
	this->atmosphere = d_atmosphere;
}

void World::setTerrain(VI_Terrain *terrain) {
	TerrainEngine *d_terrain = static_cast<TerrainEngine *>(terrain); // cast to the internal representation
	this->terrain = d_terrain;
}

void World::removeAllObjects() {
	SceneGraphNode *root = this->getRootNode();
	for(int i=0;i<root->getNChildren();i++) {
		SceneGraphNode *child = dynamic_cast<SceneGraphNode *>(root->getChild(i));
		if( child != this->viewPoint )
			child->kill();
	}
}

/*void World::checkUpdateNode(SceneGraphNode *node) {
	if( node->requires_update ) {
		//this->updateNodes->addIfAbsent(node);
		addIfAbsent_vec(this->updateNodes, node);
	}
	else {
		//this->updateNodes->remove(node);
		remove_vec(this->updateNodes, node);
	}
}

void World::removeUpdateNode(SceneGraphNode *node) {
	//this->updateNodes->remove(node);
	remove_vec(this->updateNodes, node);
	for(int i=0;i<node->getNChildren();i++) {
		SceneGraphNode *child = node->getChildAt(i);
		this->removeUpdateNode(child);
	}
}*/

/*void World::addShadowNode(SceneGraphNode *node) {
	//this->shadowNodes->add(node);
	this->shadowNodes->push_back(node);
}

void World::removeShadowNode(SceneGraphNode *node) {
	//this->shadowNodes->remove(node);
	remove_vec(this->shadowNodes, node);
	for(int i=0;i<node->getNChildren();i++) {
		SceneGraphNode *child = node->getChildAt(i);
		this->removeShadowNode(child);
	}
}*/

/*void World::noMoreChildrenAll() {
if(rootObject!=NULL)
rootObject->noMoreChildren_Rec(false);
}*/

void World::update(int time) {
	/*if(time > 1000.0)
	time = 1000.0;*/ 
	if(getRootNode() != NULL) {
		getRootNode()->update(time, true);
		/*for(int i=0;i<updateNodes->size();i++) {
		SceneGraphNode *node = static_cast<SceneGraphNode *>(updateNodes->get(i));
		// don't update children!
		node->update(time, false);
		}*/
	}

	if(viewPoint->getParent() == NULL) {
		// viewpoint can't be in the object hierarchy
		viewPoint->update(time, true);
	}
	/*if(rootObject!=NULL)
	rootObject->doKills();*/
	if(terrain != NULL)
		terrain->update(time);
	/*if(bsptree != NULL)
	bsptree->update(time);*/ // disabled for now
}

void World::init() {
}

void World::getAmbience(float *r,float *g,float *b) {
	*r = this->lightAmbient.getRf();
	*g = this->lightAmbient.getGf();
	*b = this->lightAmbient.getBf();
}

void World::setAmbiencei(unsigned char r, unsigned char g, unsigned char b) {
	this->lightAmbient.seti(r, g, b);
}

void World::setViewpoint(VI_SceneGraphNode *viewpoint) {
	SceneGraphNode *d_viewpoint = dynamic_cast<SceneGraphNode *>(viewpoint); // cast to the internal representation
	this->viewPoint = d_viewpoint;
}
